# -*- coding: utf-8 -*-

from girder import events, plugin
from girder.constants import AccessType
from girder.models.item import Item
from girder.models.folder import Folder
from .api.app import App
from . import constants, utilities

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


def _onDownloadFileComplete(event):
    item = Item().load(event.info['file']['itemId'], level=AccessType.READ)
    meta = item['meta']
    if utilities.isSlicerPackages(item):
        item_folder = Folder().load(item['folderId'], level=AccessType.READ)
        release = Folder().load(item_folder['parentId'], level=AccessType.READ)
        if item_folder['name'] == constants.EXTENSIONS_FOLDER_NAME:
            release = Folder().load(release['parentId'], level=AccessType.READ)
            if release['name'] == constants.DRAFT_RELEASE_NAME:
                Folder().increment(
                    query={'_id': release['_id']},
                    field='meta.downloadStats.%s.%s.%s.%s.%s' % (
                        meta['app_revision'], constants.EXTENSIONS_FOLDER_NAME,
                        meta['baseName'], meta['os'], meta['arch']),
                    amount=1)
            else:
                Folder().increment(
                    query={'_id': item_folder['parentId']},
                    field='meta.downloadStats.%s.%s.%s.%s' %
                          (constants.EXTENSIONS_FOLDER_NAME, meta['baseName'],
                           meta['os'], meta['arch']),
                    amount=1)
        else:
            if release['name'] == constants.DRAFT_RELEASE_NAME:
                Folder().increment(
                    query={'_id': release['_id']},
                    field='meta.downloadStats.%s.%s.%s.%s' % (
                        meta['revision'], 'applications', meta['os'], meta['arch']),
                    amount=1)
            else:
                Folder().increment(
                    query={'_id': item_folder['_id']},
                    field='meta.downloadStats.%s.%s.%s' % (
                        'applications', meta['os'], meta['arch']),
                    amount=1)


class GirderPlugin(plugin.GirderPlugin):
    DISPLAY_NAME = 'Slicer Package Manager'

    def load(self, info):
        # add plugin loading logic here
        info['apiRoot'].app = App()
        info['serverRoot'].updateHtmlVars({'title': 'Slicer package manager'})

        # Download statistics
        events.bind('model.file.download.complete', 'slicer_package_manager', _onDownloadFileComplete)

        # Mongo indexes
        Item().ensureIndex('meta.baseName')
        Item().ensureIndex('meta.os')
        Item().ensureIndex('meta.arch')
        Item().ensureIndex('meta.app_revision')
        Item().ensureIndex('updated')
        Folder().ensureIndex('meta.downloadExtensions')



# -*- coding: utf-8 -*-

import click
import platform
from tabulate import tabulate

from girder_client import GirderClient
from . import SlicerPackageManagerError, SlicerPackageClient, __version__, Constant

w = Constant.WIDTH


# ---------------- UTILITIES ---------------- #

def _getOs():
    os = platform.system()
    if os == 'Linux':
        return 'linux'
    elif os == 'Darwin':
        return 'macosx'
    elif os == 'Windows':
        return 'win'

# ------------------------------------------- #


class SlicerPackageCli(SlicerPackageClient):
    """
    A command line Python client for interacting with a Girder instance's
    RESTful api, specifically for performing uploads into a Girder instance.
    """

    def __init__(self, username, password, host=None, port=None, apiRoot=None,
                 scheme=None, apiUrl=None, apiKey=None):
        """
        Initialization function to create a SlicerPackageCli instance, will attempt
        to authenticate with the designated Girder instance. Aside from username, password,
        apiKey, and sslVerify, all other kwargs are passed directly through to the
        :py:class:`girder_client.GirderClient` base class constructor.

        :param username: username to authenticate to Girder instance.
        :param password: password to authenticate to Girder instance, leave
            this blank to be prompted.
        """
        def _progressBar(*args, **kwargs):
            bar = click.progressbar(*args, **kwargs)
            bar.bar_template = "[%(bar)s]  %(info)s  %(label)s"
            bar.show_percent = True
            bar.show_pos = True
            return bar

        super(SlicerPackageCli, self).__init__(host=host, port=port, apiRoot=apiRoot, scheme=scheme, apiUrl=apiUrl,
                                               progressReporterCls=_progressBar)
        interactive = password is None
        if apiKey:
            self.authenticate(apiKey=apiKey)
        elif username:
            self.authenticate(username, password, interactive=interactive)

    def _requestFunc(self, *args, **kwargs):
        return super(SlicerPackageCli, self)._requestFunc(*args, **kwargs)


class _HiddenOption(click.Option):
    def get_help_record(self, ctx):
        pass


class _AdvancedOption(click.Option):
    pass


class _Group(click.Group):
    def format_options(self, ctx, formatter):
        opts = []
        advanced_opts = []
        for param in self.get_params(ctx):
            rv = param.get_help_record(ctx)
            if rv is None:
                continue
            if isinstance(param, _AdvancedOption):
                advanced_opts.append(rv)
            else:
                opts.append(rv)

        if opts:
            with formatter.section('Options'):
                formatter.write_dl(opts)
        if advanced_opts:
            with formatter.section('Advanced Options'):
                formatter.write_dl(advanced_opts)
        self.format_commands(ctx, formatter)


_CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=_CONTEXT_SETTINGS)
@click.option('--api-url', default=None,
              help='RESTful API URL '
                   '(e.g https://girder.example.com:443/%s)' % GirderClient.DEFAULT_API_ROOT)
@click.option('--api-key', envvar='GIRDER_API_KEY', default=None,
              help='[default: GIRDER_API_KEY env. variable]')
@click.option('--username', default=None)
@click.option('--password', default=None)
# Advanced options
@click.option('--host', default=None,
              cls=_AdvancedOption,
              help="[default: %s]" % GirderClient.DEFAULT_HOST)
@click.option('--scheme', default=None,
              cls=_AdvancedOption,
              help="[default: %s if %s else %s]" % (
                  GirderClient.getDefaultScheme(GirderClient.DEFAULT_HOST),
                  GirderClient.DEFAULT_HOST,
                  GirderClient.getDefaultScheme("girder.example.com")))
@click.option('--port', default=None,
              cls=_AdvancedOption,
              help="[default: %s if %s; %s if %s else %s]" % (
                  GirderClient.DEFAULT_HTTPS_PORT, "https",
                  GirderClient.DEFAULT_LOCALHOST_PORT, "localhost",
                  GirderClient.DEFAULT_HTTP_PORT,
                  ))
@click.option('--api-root', default=None,
              help='relative path to the Girder REST API '
                   '[default: %s]' % GirderClient.DEFAULT_API_ROOT,
              show_default=True,
              cls=_AdvancedOption)
@click.option('--no-ssl-verify', is_flag=True, default=False,
              help='Disable SSL Verification',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--certificate', default=None,
              help='Specify path to SSL certificate',
              show_default=True,
              cls=_AdvancedOption)
@click.version_option(version=__version__, prog_name='Girder command line interface')
@click.pass_context
def main(ctx, username, password,
         api_key, api_url, scheme, host, port, api_root,
         no_ssl_verify, certificate):
    """
    The recommended way to use credentials is to first generate an API key
    and then specify the ``api-key`` argument or set the ``GIRDER_API_KEY``
    environment variable.

    The client also supports ``username`` and ``password`` args. If only the
    ``username`` is specified, the client will prompt the user to interactively
    input his/her password.
    """
    # --api-url and URL by part arguments are mutually exclusive
    url_part_options = ['host', 'scheme', 'port', 'api_root']
    has_api_url = ctx.params.get('api_url', None)
    for name in url_part_options:
        has_url_part = ctx.params.get(name, None)
        if has_api_url and has_url_part:
            raise click.BadArgumentUsage(
                'Option "--api-url" and option "--%s" are mutually exclusive.' %
                name.replace("_", "-"))
    if certificate and no_ssl_verify:
        raise click.BadArgumentUsage(
            'Option "--no-ssl-verify" and option "--certificate" are mutually exclusive.')

    ctx.obj = SlicerPackageCli(
        username, password, host=host, port=port, apiRoot=api_root,
        scheme=scheme, apiUrl=api_url, apiKey=api_key)

    if certificate and ctx.obj.scheme != 'https':
        raise click.BadArgumentUsage(
            'A URI scheme of "https" is required for option "--certificate"')


@main.group(context_settings=_CONTEXT_SETTINGS)
@click.pass_obj
def app(sc):
    pass


@main.group(context_settings=_CONTEXT_SETTINGS)
@click.pass_obj
def release(sc):
    pass


@main.group(context_settings=_CONTEXT_SETTINGS)
@click.pass_obj
def draft(sc):
    pass


@main.group(context_settings=_CONTEXT_SETTINGS)
@click.pass_obj
def extension(sc):
    pass


@main.group(context_settings=_CONTEXT_SETTINGS)
@click.pass_obj
def package(sc):
    pass


@app.command('create')
@click.argument('name')
@click.option('--desc', default=None,
              help='Description of the application',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--coll_name', default=None,
              help='Name of the new collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--coll_desc', default=None,
              help='Description of the new collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--public/--private', is_flag=True, default=None,
              help='Whether the collection should be publicly visible [default: public]',
              show_default=False,
              cls=_AdvancedOption)
@click.pass_obj
def _cli_createApp(sc, *args, **kwargs):
    """
    Create a new application.
    """
    try:
        application = sc.createApp(*args, **kwargs)
        print('%s (%s) %s' % (application['_id'], application['name'], 'CREATED'))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@app.command('list')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--name', default=None,
              help='Name of the application',
              cls=_AdvancedOption)
@click.pass_obj
def _cli_listApp(sc, *args, **kwargs):
    """
    List all the applications.
    """
    applications = sc.listApp(*args, **kwargs)
    table = []
    for application in applications:
        table.append([application['name'], application['_id']])
    print(tabulate(
        table,
        headers=['NAME', 'APPLICATION ID'],
        tablefmt="simple", numalign="left"))


@app.command('delete')
@click.argument('name')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.pass_obj
def _cli_deleteApp(sc, *args, **kwargs):
    """
    Delete an application.
    """
    try:
        application = sc.deleteApp(*args, **kwargs)
        print('%s (%s) %s' % (application['name'], application['_id'], 'DELETED'))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@release.command('create')
@click.argument('app_name', required=True)
@click.argument('name', required=True)
@click.argument('revision', required=True)
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--desc', default=None,
              help='Description of the release',
              cls=_AdvancedOption)
@click.pass_obj
def _cli_createRelease(sc, *args, **kwargs):
    """
    Create a new release.
    """
    try:
        rls = sc.createRelease(*args, **kwargs)
        print('%s %s (%s) %s' % (rls['name'], rls['meta']['revision'], rls['_id'], 'CREATED'))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@release.command('list')
@click.argument('app_name')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.pass_obj
def _cli_listRelease(sc, *args, **kwargs):
    """
    List all the releases within an application.
    """
    try:
        releases = sc.listRelease(*args, **kwargs)
        table = []
        for rls in releases:
            if 'meta' in rls and 'revision' in rls['meta']:
                revision = rls['meta']['revision']
            else:
                revision = ''
            table.append([revision, rls['name'], rls['_id']])
        print(tabulate(
            table,
            headers=['APP REVISION', 'NAME', 'RELEASE ID'],
            tablefmt="simple", numalign="left"))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@release.command('delete')
@click.argument('app_name')
@click.argument('name')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.pass_obj
def _cli_deleteRelease(sc, *args, **kwargs):
    """
    Delete a release.
    """
    try:
        rls = sc.deleteRelease(*args, **kwargs)
        print('%s %s (%s) %s' % (rls['name'], rls['meta']['revision'], rls['_id'], 'DELETED'))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@draft.command('list')
@click.argument('app_name')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--revision', default=None,
              help='Revision of the draft release',
              cls=_AdvancedOption)
@click.option('--limit', default=Constant.DEFAULT_LIMIT,
              help='The limit number of listed releases',
              cls=_AdvancedOption)
@click.option('--offset', default=0,
              help='Offset of the list',
              cls=_AdvancedOption)
@click.pass_obj
def _cli_listDraftRelease(sc, *args, **kwargs):
    """
    List all the revisions of the default preview within an application.
    """
    try:
        releases = sc.listDraftRelease(*args, **kwargs)
        table = []
        for rls in releases:
            if 'meta' in rls and 'revision' in rls['meta']:
                revision = rls['meta']['revision']
            else:
                revision = ''
            table.append([revision, rls['name'], rls['_id']])
        print(tabulate(
            table,
            headers=['APP REVISION', 'NAME', 'RELEASE ID'],
            tablefmt="simple", numalign="left"))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@draft.command('delete')
@click.argument('app_name')
@click.argument('revision')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.pass_obj
def _cli_deleteDraftRelease(sc, *args, **kwargs):
    """
    Delete a specific revision within the Draft release.
    """
    try:
        rls = sc.deleteDraftRelease(*args, **kwargs)
        print('%s %s (%s) %s' % (rls['name'], rls['meta']['revision'], rls['_id'], 'DELETED'))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@extension.command('upload')
@click.argument('app_name')
@click.argument('filepath')
@click.option('--os', 'ext_os', default=_getOs(),
              help='The target operating system of the package',
              cls=_AdvancedOption)
@click.option('--arch', default='amd64',
              help='Architecture that is supported by the extension',
              cls=_AdvancedOption)
@click.option('--name', prompt=True,
              help='The baseName of the extension',
              cls=_AdvancedOption)
@click.option('--repo_type', default='',
              help='Type of the repository where find the extension',
              cls=_AdvancedOption)
@click.option('--repo_url', default='',
              help='URL of the repository where find the extension',
              cls=_AdvancedOption)
@click.option('--revision', default='0.0.1',
              help='Revision of the extension',
              cls=_AdvancedOption)
@click.option('--app_revision', prompt=True,
              help='Revision of the application',
              cls=_AdvancedOption)
@click.option('--desc', default='',
              help='Description of the extension',
              cls=_AdvancedOption)
@click.option('--packagetype', default='',
              help='Type of the package (Installer, data...)',
              cls=_AdvancedOption)
@click.option('--icon_url', default='',
              help="Url of the extension's logo",
              cls=_AdvancedOption)
@click.option('--category', default=None,
              help='Category of the extension',
              cls=_AdvancedOption)
@click.option('--homepage', default='',
              help='Url of the extension homepage',
              cls=_AdvancedOption)
@click.option('--screenshots', default=None,
              help='Space-separate list of URLs of screenshots for the extension.',
              cls=_AdvancedOption)
@click.option('--contributors', default=None,
              help='List of contributors of the extension.',
              cls=_AdvancedOption)
@click.option('--dependency', default=None,
              help='List of the required extensions to use this one.',
              cls=_AdvancedOption)
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--force', default=False,
              help='Force the upload',
              cls=_AdvancedOption)
@click.pass_obj
def _cli_uploadExtension(sc, *args, **kwargs):
    """
    Upload an extension.
    """
    try:
        print('Create the extension %s' % kwargs['name'])
        ext = sc.uploadExtension(*args, **kwargs)
        if ext == Constant.EXTENSION_AREADY_UP_TO_DATE:
            print('Extension "%s" is already up-to-date\t(Extension Item updated)' % kwargs['name'])
        elif ext == Constant.EXTENSION_NOW_UP_TO_DATE:
            print('%s %s %s' % (kwargs['name'], 'UPLOADED', 'The extension is now up-to-date'))
        else:
            print('%s (%s) %s' % (ext['name'], ext['_id'], 'UPLOADED'))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@extension.command('download')
@click.argument('app_name')
@click.argument('id_or_name')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--dir_path', default=Constant.CURRENT_FOLDER,
              help='Path to the directory where will be downloaded the extension',
              cls=_AdvancedOption)
@click.pass_obj
def _cli_downloadExtension(sc, *args, **kwargs):
    """
    Download an extension.
    """
    try:
        print('Start download...')
        ext = sc.downloadExtension(*args, **kwargs)
        print('%s (%s) %s [%s]' % (ext['name'], ext['_id'], 'DOWNLOADED', kwargs['dir_path']))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@extension.command('list')
@click.argument('app_name')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--name', default=None,
              help='The baseName of the extension',
              cls=_AdvancedOption)
@click.option('--os', 'ext_os', type=click.Choice(['win', 'linux', 'macosx']))
@click.option('--arch', type=click.Choice(['amd64', 'i386']))
@click.option('--app_revision', default=None,
              help='The revision of the application',
              cls=_AdvancedOption)
@click.option('--release', default=Constant.DRAFT_RELEASE_NAME,
              help='List all extension within the release',
              cls=_AdvancedOption)
@click.option('--limit', default=Constant.DEFAULT_LIMIT,
              help='The limit number of listed extensions ',
              cls=_AdvancedOption)
@click.option('--all', is_flag=True,
              default=False,
              help='List all the extension of the application',
              cls=_AdvancedOption)
@click.pass_obj
def _cli_listExtension(sc, *args, **kwargs):
    """
    List all the extensions within an application.
    """
    try:
        extensions = sc.listExtension(*args, **kwargs)
        rls_list = sc.listRelease(app_name=kwargs['app_name'], coll_id=kwargs['coll_id'])
        table = []
        for ext in extensions:
            release_name = None
            for rls in rls_list:
                if rls['meta']['revision'] == ext['meta']['app_revision']:
                    release_name = rls['name']
                    break
            if not release_name:
                release_name = Constant.DRAFT_RELEASE_NAME
            table.append([ext['meta']['revision'], ext['name'], release_name,
                          ext['meta']['app_revision'], ext['_id']])
        print(tabulate(
            table,
            headers=['REVISION', 'NAME', 'RELEASE NAME', 'APP REVISION', 'EXTENSION ID'],
            tablefmt="simple", numalign="left"))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@extension.command('delete')
@click.argument('app_name')
@click.argument('id_or_name')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.pass_obj
def _cli_deleteExtension(sc, *args, **kwargs):
    """
    Delete an extension by ID or Name.
    """
    try:
        ext = sc.deleteExtension(*args, **kwargs)
        print('%s %s (%s) %s' % (ext['name'], ext['meta']['revision'], ext['_id'], 'DELETED'))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@package.command('upload')
@click.argument('app_name')
@click.argument('filepath')
@click.option('--os', 'pkg_os', default=_getOs(),
              help='The target operating system of the package',
              cls=_AdvancedOption)
@click.option('--arch', default='amd64',
              help='Architecture that is supported by the package',
              cls=_AdvancedOption)
@click.option('--name', prompt=True,
              help='The baseName of the package',
              cls=_AdvancedOption)
@click.option('--repo_type', default='',
              help='Type of the repository where find the package',
              cls=_AdvancedOption)
@click.option('--repo_url', default='',
              help='URL of the repository where find the package',
              cls=_AdvancedOption)
@click.option('--revision', prompt=True,
              help='Revision of the application',
              cls=_AdvancedOption)
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--desc', default='',
              help='Description of the package',
              cls=_AdvancedOption)
@click.option('--pre_release', is_flag=True, default=None,
              help='Boolean to specify if the package is ready to be distributed',
              cls=_AdvancedOption)
@click.pass_obj
def _cli_uploadApplicationPackage(sc, *args, **kwargs):
    """
    Upload an application package.
    """
    try:
        print('Create the application package %s' % kwargs['name'])
        pkg = sc.uploadApplicationPackage(*args, **kwargs)
        if pkg == Constant.PACKAGE_NOW_UP_TO_DATE:
            print('%s %s %s' % (kwargs['name'], 'UPLOADED', 'The package is now up-to-date'))
        else:
            print('%s (%s) %s' % (pkg['name'], pkg['_id'], 'UPLOADED'))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@package.command('download')
@click.argument('app_name')
@click.argument('id_or_name')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--dir_path', default=Constant.CURRENT_FOLDER,
              help='Path to the directory where will be downloaded the package',
              cls=_AdvancedOption)
@click.pass_obj
def _cli_downloadApplicationPackage(sc, *args, **kwargs):
    """
    Download an application package.
    """
    try:
        print('Start download...')
        pkg = sc.downloadApplicationPackage(*args, **kwargs)
        print('%s (%s) %s [%s]' % (pkg['name'], pkg['_id'], 'DOWNLOADED', kwargs['dir_path']))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@package.command('list')
@click.argument('app_name')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.option('--name', default=None,
              help='The baseName of the package',
              cls=_AdvancedOption)
@click.option('--os', 'pkg_os', type=click.Choice(['win', 'linux', 'macosx']))
@click.option('--arch', type=click.Choice(['amd64', 'i386']))
@click.option('--revision', default=None,
              help='The revision of the application',
              cls=_AdvancedOption)
@click.option('--release', default=None,
              help='List all packages within the release',
              cls=_AdvancedOption)
@click.option('--limit', default=Constant.DEFAULT_LIMIT,
              help='The limit number of listed packages ',
              cls=_AdvancedOption)
@click.pass_obj
def _cli_listApplicationPackage(sc, *args, **kwargs):
    """
    List all the application packages within an application.
    """
    try:
        packages = sc.listApplicationPackage(*args, **kwargs)
        rls_list = sc.listRelease(app_name=kwargs['app_name'], coll_id=kwargs['coll_id'])
        table = []
        for pkg in packages:
            release_name = None
            for rls in rls_list:
                if rls['meta']['revision'] == pkg['meta']['revision']:
                    release_name = rls['name']
                    break
            if not release_name:
                release_name = Constant.DRAFT_RELEASE_NAME
            table.append([pkg['meta']['revision'], pkg['name'], release_name, pkg['_id']])
        print(tabulate(
            table,
            headers=['APP REVISION', 'NAME', 'RELEASE NAME', 'PACKAGE ID'],
            tablefmt="simple", numalign="left"))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)


@package.command('delete')
@click.argument('app_name')
@click.argument('id_or_name')
@click.option('--coll_id', default=None, envvar='COLLECTION_ID',
              help='ID of an existing collection',
              show_default=True,
              cls=_AdvancedOption)
@click.pass_obj
def _cli_deleteApplicationPackage(sc, *args, **kwargs):
    """
    Delete an application package by ID or Name.
    """
    try:
        pkg = sc.deleteApplicationPackage(*args, **kwargs)
        print('%s %s (%s) %s' % (pkg['name'], pkg['meta']['revision'], pkg['_id'], 'DELETED'))
    except SlicerPackageManagerError as exc_info:
        print(exc_info)



import os


FIXTURE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'data',
    )


extensions = {
    'extension1': {
        'meta': {
            'os': 'linux',
            'arch': 'i386',
            'baseName': 'Ext1',
            'repository_type': 'git',
            'repository_url': 'http://slicer.com/extension/Ext',
            'revision': '35333',
            'app_revision': '0005',
            'packagetype': 'installer',
            'codebase': 'SL4',
            'description': 'Extension for Slicer 4'
        }
    },
    'extension2': {
        'meta': {
            'os': 'win',
            'arch': 'i386',
            'baseName': 'Ext2',
            'repository_type': 'git',
            'repository_url': 'http://slicer.com/extension/Ext',
            'revision': '54342',
            'app_revision': '0000',
            'packagetype': 'installer',
            'codebase': 'SL4',
            'description': 'Extension for Slicer 4 new version'
        }
    },
    'extension3': {
        'meta': {
            'os': 'linux',
            'arch': 'amd64',
            'baseName': 'Ext3',
            'repository_type': 'gitlab',
            'repository_url': 'http://slicer.com/extension/Ext',
            'revision': '542',
            'app_revision': '0001',
            'packagetype': 'zip',
            'codebase': 'SL434334',
            'description': 'Extension for Slicer 4 new version'
        }
    },
    'extension4': {
        'meta': {
            'os': 'macosx',
            'arch': 'amd64',
            'baseName': 'Ext3',
            'repository_type': 'gitlab',
            'repository_url': 'http://slicer.com/extension/Ext',
            'revision': '542',
            'app_revision': '0001',
            'packagetype': 'zip',
            'codebase': 'SL434334',
            'description': 'Extension for Slicer 4 new version'
        }
    },
    'extension5': {
        'meta': {
            'os': 'macosx',
            'arch': 'i386',
            'baseName': 'Ext3',
            'repository_type': 'gitlab',
            'repository_url': 'http://slicer.com/extension/Ext',
            'revision': '542',
            'app_revision': '0001',
            'packagetype': 'zip',
            'codebase': 'SL434334',
            'description': 'Extension for Slicer 4 new version'
        }
    }
}

packages = {
    'package1': {
        'meta': {
            'os': 'macosx',
            'arch': 'amd64',
            'baseName': 'pkg1',
            'repository_type': 'gitlab',
            'repository_url': 'https://slicer4.com',
            'revision': '0005',
        }
    },

    'package2': {
        'meta': {
            'os': 'win',
            'arch': 'i386',
            'baseName': 'pkg2',
            'repository_type': 'gitlab',
            'repository_url': 'https://slicer4.com',
            'revision': '0000',
        }
    },

    'package3': {
        'meta': {
            'os': 'linux',
            'arch': 'amd64',
            'baseName': 'pkg3',
            'repository_type': 'git',
            'repository_url': 'git://slicer4.com',
            'revision': '0000',
        }
    }

}

expectedDownloadStats = {
    '0000': {
        'applications': {
            'win': {
                'i386': 1
            },
            'linux': {
                'amd64': 1
            }
        },
        'extensions': {
            'Ext2': {
                'win': {
                    'i386': 1
                }
            }
        }
    },
    '0001': {
        'extensions': {
            'Ext3': {
                'linux': {
                    'amd64': 1
                },
                'macosx': {
                    'amd64': 1,
                    'i386': 1
                }
            }
        }
    },
    '0005': {
        'applications': {
            'macosx': {
                'amd64': 1
            }
        },
        'extensions': {
            'Ext1': {
                'linux': {
                    'i386': 1
                }
            }
        }
    }
}


def computeContentChecksum(algo, content):
    """Compute digest of ``content`` using ``algo``.

    Supported hashing algorithms are SHA256, SHA512, and MD5.

    :raises ValueError: if algo is unknown.
    """
    import hashlib

    if algo not in ['SHA256', 'SHA512', 'MD5']:
        raise ValueError("unsupported hashing algorithm %s" % algo)

    hash = hashlib.new(algo)
    hash.update(content)
    return hash.hexdigest()


def computeFileChecksum(algo, filePath):
    """Compute digest of ``filePath`` using ``algo``.

    Supported hashing algorithms are SHA256, SHA512, and MD5.
    It internally reads the file by chunk of 8192 bytes.

    :raises ValueError: if algo is unknown.
    :raises IOError: if filePath does not exist.
    """
    import hashlib

    if algo not in ['SHA256', 'SHA512', 'MD5']:
        raise ValueError("unsupported hashing algorithm %s" % algo)

    with open(filePath, 'rb') as content:
        hash = hashlib.new(algo)
        while True:
            chunk = content.read(8192)
            if not chunk:
                break
            hash.update(chunk)
        return hash.hexdigest()


class ExternalData:

    def __init__(self, objectStorePath):
        self._downloadPercent = 0
        self._objectStorePath = objectStorePath

    @staticmethod
    def _humanFormatSize(size):
        """Convert size in bytes to human readable string including unit suffix.

        Adapted from
        https://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
        """
        for x in ['bytes', 'KB', 'MB', 'GB']:
            if -1024.0 < size < 1024.0:
                return "%3.1f %s" % (size, x)
            size /= 1024.0
        return "%3.1f %s" % (size, 'TB')

    @staticmethod
    def _extractAlgoAndDigest(checksum):
        """Given a checksum string formatted as ``<algo>:<digest>`` returns the tuple ``(algo, digest)``.

        ``<algo>`` is expected to be `SHA256`, `SHA512`, or `MD5`.
        ``<digest>`` is expected to be the full length hexdecimal digest.

        :raises ValueError: if checksum is incorrectly formatted.
        """
        if checksum is None:
            return None, None
        if len(checksum.split(':')) != 2:
            raise ValueError("invalid checksum '%s'. Expected format is '<algo>:<digest>'." % checksum)
        (algo, digest) = checksum.split(':')
        expected_algos = ['SHA256', 'SHA512', 'MD5']
        if algo not in expected_algos:
            raise ValueError("invalid algo '%s'. Algo must be one of %s" % (algo, ", ".join(expected_algos)))
        expected_digest_length = {'SHA256': 64, 'SHA512': 128, 'MD5': 32}
        if len(digest) != expected_digest_length[algo]:
            raise ValueError("invalid digest length %d. Expected digest length for %s is %d" % (
                len(digest), algo, expected_digest_length[algo]))
        return algo, digest

    def _reportHook(self, blocksSoFar, blockSize, totalSize):
        # we clamp to 100% because the blockSize might be larger than the file itself
        percent = min(int((100. * blocksSoFar * blockSize) / totalSize), 100)
        if percent == 100 or (percent - self._downloadPercent >= 10):
            # we clamp to totalSize when blockSize is larger than totalSize
            humanSizeSoFar = self._humanFormatSize(min(blocksSoFar * blockSize, totalSize))
            humanSizeTotal = self._humanFormatSize(totalSize)
            print('Downloaded %s (%d%% of %s)...' % (humanSizeSoFar, percent, humanSizeTotal))
            self._downloadPercent = percent

    def _downloadFile(self, uri, name, checksum):
        """
        :param uri: Download URL.
        :param name: File name that will be downloaded.
        :param checksum: Checksum formatted as ``<algo>:<digest>`` to verify the downloaded file.
            For example, ``SHA256:cc211f0dfd9a05ca3841ce1141b292898b2dd2d3f08286affadf823a7e58df93``.
        """
        self._downloadPercent = 0
        filePath = self._objectStorePath + '/' + name
        (algo, digest) = self._extractAlgoAndDigest(checksum)
        if not os.path.exists(filePath) or os.stat(filePath).st_size == 0:
            import urllib.error
            import urllib.parse
            import urllib.request
            print('Requesting download %s from %s ...' % (name, uri))
            try:
                urllib.request.urlretrieve(uri, filePath, self._reportHook)
                print('Download finished')
            except IOError as exc:
                raise ValueError(f"Failed to download {uri} to {filePath}: %s" % exc)

            if algo is not None:
                print('Verifying checksum')
                current_digest = computeFileChecksum(algo, filePath)
                if current_digest != digest:
                    print(
                        'Checksum verification failed. Computed checksum %s different from expected checksum %s' % (
                            current_digest, digest))
                    os.remove(filePath)
                else:
                    self._downloadPercent = 100
                    print('Checksum OK')
        else:
            if algo is not None:
                print('Verifying checksum')
                current_digest = computeFileChecksum(algo, filePath)
                if current_digest != digest:
                    print('File already exists in cache but checksum is different - re-downloading it.')
                    os.remove(filePath)
                    return self.downloadFile(uri, self._objectStorePath, name, checksum)
                else:
                    self._downloadPercent = 100
                    print('File already exists and checksum is OK - reusing it.')
            else:
                self._downloadPercent = 100
                print('File already exists in cache - reusing it.')
        return filePath

    def download(self, uri, fileName, checksum):
        """Download data into storage directory.

        :param uri: Download URL.
        :param fileName: File name that will be downloaded.
        :param checksum: Checksum formatted as ``<algo>:<digest>`` to verify the downloaded file.
            For example, ``SHA256:cc211f0dfd9a05ca3841ce1141b292898b2dd2d3f08286affadf823a7e58df93``.
        """
        if not os.access(self._objectStorePath, os.W_OK):
            os.makedirs(self._objectStorePath, exist_ok=True)
            assert os.access(self._objectStorePath, os.W_OK)
        maximumAttemptsCount = 3
        errors = set()
        for _attemptsCount in range(maximumAttemptsCount):
            try:
                return self._downloadFile(uri, fileName, checksum)
            except ValueError as exc:
                errors.add(str(exc))
                continue
        raise RuntimeError('Download of %s failed for %d attempts\n  uri: %s\n  errors: %s' % (
            fileName, maximumAttemptsCount, uri, ", ".join(errors)
        ))



import os

import pytest

from shutil import copyfile

from . import ExternalData


@pytest.fixture
def external_data(request, tmpdir):
    """A pytest fixture to define a 'tmpdir' containing files or directories
    specified with a 'external_data' mark.
    """
    entry_list = []
    for mark in request.node.iter_markers('external_data'):
        entry_list.extend(mark.args)

    # Collect externals in map where keys are "fileName" and values are "(<algo>, <checksum>)"
    externals = {}
    for entry in entry_list:
        ext_not_found = []
        for ext in ['sha256', 'sha512', 'md5']:
            external = entry + "." + ext
            if os.path.exists(external):
                break
            else:
                ext_not_found.append(ext)
                external = None
        if external is None:
            raise ValueError("{}.{} not found".format(entry, ", ".join(ext_not_found)))
        with open(external) as content:
            hashsum = content.read().strip()
        externals[os.path.basename(entry)] = (ext, hashsum)

    external_data_dir = os.environ.get(
        'GIRDER_TEST_DATA_PREFIX', os.path.join(os.path.dirname(__file__), ".external_data"))

    # Download files
    downloaded_files = []
    for fileName, value in externals.items():
        algo, checksum = value
        print("fileName [%s] algo [%s] checksum [%s]" % (fileName, algo, checksum))
        downloaded_files.append(
            ExternalData(external_data_dir).download(
                uri="https://data.kitware.com/api/v1/file/hashsum/{}/{}/download".format(algo, checksum),
                fileName=fileName,
                checksum="{}:{}".format(algo.upper(), checksum)
            )
        )

    # Copy download files to temporary directory
    for downloaded_file in downloaded_files:
        copyfile(downloaded_file, tmpdir.join(os.path.basename(downloaded_file)))

    return tmpdir



# -*- coding: utf-8 -*-

import logging
import os
import pytest
import re

from click.testing import CliRunner
from slicer_package_manager_client.cli import main


APPS = ['App', 'App1', 'App2']

RELEASES = [
    {
        'app_name': APPS[0],
        'name': 'Release',
        'revision': 'r000'
    },
    {
        'app_name': APPS[0],
        'name': 'Release1',
        'revision': 'r001'
    }
]

DRAFT_RELEASES = [
    {
        'revision': 'r002'
    },
    {
        'revision': 'r003'
    }
]

PACKAGES = [
    {
        'filepath': './file1.txt',
        'app_name': APPS[0],
        'os': 'macosx',
        'arch': 'i386',
        'revision': DRAFT_RELEASES[0]['revision'],
        'repo_type': 'git',
        'repo_url': 'git@github.com:pkg1.git',
        'baseName': 'pkg1'
    },
    {
        'filepath': './file2.txt',
        'app_name': APPS[0],
        'os': 'macosx',
        'arch': 'amd64',
        'revision': DRAFT_RELEASES[0]['revision'],
        'repo_type': 'git',
        'repo_url': 'git@github.com:pkg2.git',
        'baseName': 'pkg2'
    },
    {
        'filepath': './file3.txt',
        'app_name': APPS[0],
        'os': 'win',
        'arch': 'i386',
        'revision': DRAFT_RELEASES[0]['revision'],
        'repo_type': 'git',
        'repo_url': 'git@github.com:pkg3.git',
        'baseName': 'pkg3'
    },
    {
        'filepath': './file4.txt',
        'app_name': APPS[0],
        'os': 'linux',
        'arch': 'amd64',
        'revision': DRAFT_RELEASES[1]['revision'],
        'repo_type': 'git',
        'repo_url': 'git@github.com:pkg4.git',
        'baseName': 'pkg4'
    }
]

EXTENSIONS = [
    {
        'filepath': './file1.txt',
        'app_name': APPS[0],
        'os': 'macosx',
        'arch': 'i386',
        'revision': '000',
        'repo_type': 'git',
        'repo_url': 'git@github.com:ext1.git',
        'baseName': 'ext1',
        'app_revision': RELEASES[0]['revision']
    },
    {
        'filepath': './file2.txt',
        'app_name': APPS[0],
        'os': 'linux',
        'arch': 'amd64',
        'revision': '001',
        'repo_type': 'git',
        'repo_url': 'git@github.com:ext2.git',
        'baseName': 'ext2',
        'app_revision': RELEASES[0]['revision']
    },
    {
        'filepath': './file3.txt',
        'app_name': APPS[0],
        'os': 'win',
        'arch': 'amd64',
        'revision': '000',
        'repo_type': 'git',
        'repo_url': 'git@github.com:ext3.git',
        'baseName': 'ext3',
        'app_revision': DRAFT_RELEASES[0]['revision']
    },
    {
        'filepath': './file4.txt',
        'app_name': APPS[0],
        'os': 'linux',
        'arch': 'amd64',
        'revision': '000',
        'repo_type': 'git',
        'repo_url': 'git@github.com:ext4.git',
        'baseName': 'ext4',
        'app_revision': DRAFT_RELEASES[1]['revision']
    }
]

CLI_COMMON_ARGS = [
    '--api-url', 'http://localhost:8080/api/v1',
    '--username', 'admin',
    '--password', 'password'
]


def _cli_runner_invoke(main, cmd):
    res = CliRunner().invoke(main, cmd)
    if res.exit_code != 0:
        logging.error(res.exception)
    return res


def _cli_upload_package(package):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['package', 'upload', package['app_name'], package['filepath'],
                '--os', package['os'],
                '--arch', package['arch'],
                '--name', package['baseName'],
                '--revision', package['revision'],
                '--repo_type', package['repo_type'],
                '--repo_url', package['repo_url']])
    return _cli_runner_invoke(main, cmd)


def _cli_upload_extension(extension):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['extension', 'upload', extension['app_name'], extension['filepath'],
                '--os', extension['os'],
                '--arch', extension['arch'],
                '--name', extension['baseName'],
                '--revision', extension['revision'],
                '--app_revision', extension['app_revision'],
                '--repo_type', extension['repo_type'],
                '--repo_url', extension['repo_url']])
    return _cli_runner_invoke(main, cmd)


@pytest.mark.vcr()
@pytest.fixture(autouse=True)
def TearDown(server):
    yield
    for idx in range(len(APPS)):
        cmd = list(CLI_COMMON_ARGS)
        cmd.extend(['app', 'delete', APPS[idx]])
        _cli_runner_invoke(main, cmd)


@pytest.mark.vcr()
@pytest.fixture
def apps(server):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['app', 'create', APPS[0]])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.match(r"\w{24} \(%s\) CREATED" % re.escape(APPS[0]), res.output)
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['app', 'create', APPS[1]])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.match(r"\w{24} \(%s\) CREATED" % re.escape(APPS[1]), res.output)
    yield


@pytest.mark.vcr()
@pytest.fixture
def releases(server):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['release', 'create', RELEASES[1]['app_name'], RELEASES[1]['name'], RELEASES[1]['revision']])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.match(r"%s %s \(\w{24}\) CREATED" % (RELEASES[1]['name'], RELEASES[1]['revision']), res.output)
    yield


@pytest.mark.vcr()
@pytest.fixture
def packages(server, apps, files):
    res = _cli_upload_package(PACKAGES[0])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % (getAppPkgName(PACKAGES[0])), res.output)

    res = _cli_upload_package(PACKAGES[1])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % (getAppPkgName(PACKAGES[1])), res.output)

    res = _cli_upload_package(PACKAGES[2])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % (getAppPkgName(PACKAGES[2])), res.output)

    res = _cli_upload_package(PACKAGES[3])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % (getAppPkgName(PACKAGES[3])), res.output)


@pytest.mark.vcr()
@pytest.fixture
def extensions(server, apps, files):
    res = _cli_upload_extension(EXTENSIONS[0])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % re.escape(getExtPkgName(EXTENSIONS[0])), res.output)

    res = _cli_upload_extension(EXTENSIONS[1])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % re.escape(getExtPkgName(EXTENSIONS[1])), res.output)

    res = _cli_upload_extension(EXTENSIONS[2])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % re.escape(getExtPkgName(EXTENSIONS[2])), res.output)

    res = _cli_upload_extension(EXTENSIONS[3])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % re.escape(getExtPkgName(EXTENSIONS[3])), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testCreateAppCLI(server):

    def _create(app):
        cmd = list(CLI_COMMON_ARGS)
        cmd.extend(['app', 'create', APPS[2]])
        return _cli_runner_invoke(main, cmd)

    res = _create(APPS[2])
    assert res.exit_code == 0
    assert re.match(r"\w{24} \(%s\) CREATED" % re.escape(APPS[2]), res.output)

    # Try to create the same app
    res = _create(APPS[2])
    assert res.exit_code == 0
    assert re.match(r"The Application \"%s\" already exist\." % re.escape(APPS[2]),
                    res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testListAppCLI(server, apps):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['app', 'list'])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.search(r"%s *\w{24}" % re.escape(APPS[0]), res.output)
    assert re.search(r"%s *\w{24}" % re.escape(APPS[1]), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDeleteAppCLI(server, apps):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['app', 'delete', APPS[0]])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.match(r"%s \(\w{24}\) DELETED" % re.escape(APPS[0]), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testCreateReleaseCLI(server, apps):

    def _create(release):
        cmd = list(CLI_COMMON_ARGS)
        cmd.extend(['release', 'create', release['app_name'], release['name'], release['revision']])
        return _cli_runner_invoke(main, cmd)

    res = _create(RELEASES[0])
    assert res.exit_code == 0
    assert re.match(r"%s %s \(\w{24}\) CREATED" % (RELEASES[0]['name'], RELEASES[0]['revision']), res.output)

    # Try to create the same release
    res = _create(RELEASES[0])
    assert res.exit_code == 0
    assert re.match(r"The release \"%s\" already exist\." % re.escape(RELEASES[0]['name']), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testListReleaseCLI(server, apps, releases):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['release', 'list', RELEASES[1]['app_name']])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.search(r"%s *%s *\w{24}" % (RELEASES[1]['revision'], RELEASES[1]['name']), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDeleteReleaseCLI(server, apps, releases):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['release', 'delete', RELEASES[1]['app_name'], RELEASES[1]['name']])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.match(r"%s %s \(\w{24}\) DELETED" % (RELEASES[1]['name'], RELEASES[1]['revision']), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testlistDraftCLI(server, packages):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['draft', 'list', APPS[0]])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.search(r"%s *%s *\w{24}" % ('r002', 'r002'), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testListDraftWithLimitCLI(server, packages):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['draft', 'list', APPS[0], '--limit', '1'])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.search(r"%s *%s *\w{24}" % ('r003', 'r003'), res.output)
    assert not re.search(r"%s *%s *\w{24}" % ('r002', 'r002'), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDeleteDraftCLI(server, packages):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['draft', 'delete', APPS[0], 'r002'])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.search(r"%s %s \(\w{24}\) DELETED" % ('r002', 'r002'), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testUploadPackagesCLI(server, apps, files):
    res = _cli_upload_package(PACKAGES[0])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % re.escape(getAppPkgName(PACKAGES[0])), res.output)

    res = _cli_upload_package(PACKAGES[1])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % re.escape(getAppPkgName(PACKAGES[1])), res.output)

    res = _cli_upload_package(PACKAGES[2])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % re.escape(getAppPkgName(PACKAGES[2])), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testListPackagesCLI(server, packages):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['package', 'list', APPS[0]])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0

    name = getAppPkgName(PACKAGES[0])
    assert re.search(r"%s *%s *%s *\w{24}" % (PACKAGES[0]['revision'], name, 'draft'), res.output)

    name = getAppPkgName(PACKAGES[1])
    assert re.search(r"%s *%s *%s *\w{24}" % (PACKAGES[1]['revision'], name, 'draft'), res.output)

    name = getAppPkgName(PACKAGES[2])
    assert re.search(r"%s *%s *%s *\w{24}" % (PACKAGES[2]['revision'], name, 'draft'), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDeletePackagesCLI(server, packages):
    cmd = list(CLI_COMMON_ARGS)
    name = getAppPkgName(PACKAGES[2])
    cmd.extend(['package', 'delete', PACKAGES[2]['app_name'], name])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.match(r"%s %s \(\w{24}\) DELETED" % (name, PACKAGES[2]['revision']), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDownloadPackagesCLI(server, packages):
    cmd = list(CLI_COMMON_ARGS)
    name = getAppPkgName(PACKAGES[0])
    cmd.extend(['package', 'download', PACKAGES[0]['app_name'], name])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) DOWNLOADED \[.*]" % name, res.output)
    os.remove('%s.txt' % name)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testUploadExtensionsCLI(server, apps, files):
    res = _cli_upload_extension(EXTENSIONS[0])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % re.escape(getExtPkgName(EXTENSIONS[0])), res.output)

    res = _cli_upload_extension(EXTENSIONS[1])
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) UPLOADED" % re.escape(getExtPkgName(EXTENSIONS[1])), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testListExtensionsCLI(server, extensions):
    cmd = list(CLI_COMMON_ARGS)
    cmd.extend(['extension', 'list', APPS[0], '--all'])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0

    name = getExtPkgName(EXTENSIONS[0])
    assert re.search(r"%s *%s *%s *%s *\w{24}" % (EXTENSIONS[0]['revision'], name, 'draft',
                                                  EXTENSIONS[0]['app_revision']), res.output)

    name = getExtPkgName(EXTENSIONS[1])
    assert re.search(r"%s *%s *%s *%s *\w{24}" % (EXTENSIONS[1]['revision'], name, 'draft',
                                                  EXTENSIONS[1]['app_revision']), res.output)

    name = getExtPkgName(EXTENSIONS[2])
    assert re.search(r"%s *%s *%s *%s *\w{24}" % (EXTENSIONS[2]['revision'], name, 'draft',
                                                  EXTENSIONS[2]['app_revision']), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDeleteExtensionsCLI(server, extensions):
    cmd = list(CLI_COMMON_ARGS)
    name = getExtPkgName(EXTENSIONS[0])
    cmd.extend(['extension', 'delete', EXTENSIONS[0]['app_name'], name])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.match(r"%s %s \(\w{24}\) DELETED" % (name, EXTENSIONS[0]['revision']), res.output)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDownloadExtensionsCLI(server, extensions):
    cmd = list(CLI_COMMON_ARGS)
    name = getExtPkgName(EXTENSIONS[0])
    cmd.extend(['extension', 'download', EXTENSIONS[0]['app_name'], name])
    res = _cli_runner_invoke(main, cmd)
    assert res.exit_code == 0
    assert re.search(r"%s \(\w{24}\) DOWNLOADED \[.*]" % name, res.output)
    os.remove('%s.txt' % name)


def getAppPkgName(package):
    return "{baseName}_{os}_{arch}_{revision}".format(**package)


def getExtPkgName(extension):
    return "{app_revision}_{baseName}_{os}_{arch}_{revision}".format(**extension)



# -*- coding: utf-8 -*-

import os
from bson.objectid import ObjectId

from girder_client import GirderClient

__version__ = '0.0.1'
__license__ = 'Apache 2.0'

appName = 'Slicer'


class Constant:
    """
    A bunch of utilities constant, as to handle ``Error`` or set default parameters.
    """

    # Success
    PACKAGE_NOW_UP_TO_DATE = 31
    EXTENSION_AREADY_UP_TO_DATE = 32
    EXTENSION_NOW_UP_TO_DATE = 33

    # Default
    CURRENT_FOLDER = os.getcwd()
    DRAFT_RELEASE_NAME = 'draft'
    DEFAULT_LIMIT = 50

    # Display
    WIDTH = 25  # Shouldn't be less than 24


class SlicerPackageManagerError(Exception):
    pass


class SlicerPackageClient(GirderClient):
    """
    The SlicerPackageClient allows to use the slicer_package_manager plugin of Girder.
    This allow to manage 5 top level entities:

        * Application
        * Release
        * Draft
        * Package
        * Extension

    It's now possible to choose the collection within create the application. It's also
    possible to get an existing collection by ID for creating the application inside.

    In this case, you must provide the ``coll_id`` argument to be able to use all the
    commands on these application. By default all the command look for application
    which are under the *Applications* collection.
    """

    def __init__(self, host=None, port=None, apiRoot=None, scheme=None, apiUrl=None,
                 progressReporterCls=None):
        super(SlicerPackageClient, self).__init__(
            host=host, port=port, apiRoot=apiRoot, scheme=scheme, apiUrl=apiUrl,
            progressReporterCls=progressReporterCls)

    def createApp(self, name, desc=None, coll_id=None, coll_name=None, coll_desc=None,
                  public=None):
        """
        Create a new application in the collection which correspond to ``coll_id``,
        by default it will create the application in the collection named ``Applications``.
        The application will contain a ``draft`` folder.
        Two templates names will be set as a metadata of this new application.
        One for determine each future uploaded application package and the
        other to determine each future uploaded extension.
        It's also possible to create a new collection by specifying "coll_name".
        If this collection already exist it will use it.

        :param name: name of the new application
        :param desc: Optional description of the application
        :param coll_id: Id of an existing collection
        :param coll_name: Name of the collection
        :param coll_desc: Optional description of the new collection
        :param public: Whether the collection should be publicly visible
        :return:  The new application
        """
        apps = self.listApp(name=name)
        if apps:
            raise SlicerPackageManagerError('The Application "%s" already exist.' % name)
        return self.post('/app', parameters={
            'name': name,
            'app_description': desc,
            'collection_id': coll_id,
            'collection_name': coll_name,
            'collection_description': coll_desc,
            'public': public
        })

    def listApp(self, name=None, coll_id=None):
        """
        List all the applications within a specific collection by providing the option
        ``coll_id``. By default it will list within the collection ``Applications``.
        It can also lead to get the application by name.

        :param name: application mame
        :param coll_id: Collection ID
        :return:  A list of applications
        """
        apps = self.get('/app', parameters={
            'collection_id': coll_id,
            'name': name
        })
        return apps

    def deleteApp(self, name, coll_id=None):
        """
        Delete the application by ID.

        :param name: application name
        :param coll_id: Collection ID
        :return: The deleted application
        """
        app = self._getApp(app_name=name, coll_id=coll_id)
        self.delete('/app/%s' % app['_id'])
        return app

    def createRelease(self, app_name, name, revision, coll_id=None, desc=None):
        """
        Create a new release within the application corresponding to ``app_name``.

        :param app_name: Name of the application
        :param name: Name of the release
        :param revision: Revision of the application
        :param coll_id: Collection ID
        :param desc: Description of the release
        :return: The new release
        """
        app = self._getApp(app_name=app_name, coll_id=coll_id)
        releases = self.listRelease(app_name=app_name, name=name)
        if releases:
            raise SlicerPackageManagerError('The release "%s" already exist.' % name)
        return self.post('/app/%s/release' % app['_id'], parameters={
            'name': name,
            'app_revision': revision,
            'description': desc
        })

    def listRelease(self, app_name, name=None, coll_id=None):
        """
        List all the release within an application. It's also able to get
        one specific release by name.

        :param app_name: Name of the application
        :param name: Name of the release
        :param coll_id: Collection ID
        :return: A list of all the release within the application
        """
        app = self._getApp(app_name=app_name, coll_id=coll_id)
        if name:
            releases = self.get(
                '/app/%s/release' % app['_id'],
                parameters={'release_id_or_name': name})
        else:
            releases = self.get('/app/%s/release' % app['_id'])
        return releases

    def deleteRelease(self, app_name, name, coll_id=None):
        """
        Delete a release within an application.

        :param app_name: Name of the application
        :param name: Name of the release
        :param coll_id: Collection ID
        :return: The deleted release
        """
        app = self._getApp(app_name=app_name, coll_id=coll_id)
        release = self.listRelease(app_name, name)
        if not release:
            raise SlicerPackageManagerError('The release "%s" doesn\'t exist.' % name)
        self.delete('/app/%s/release/%s' % (app['_id'], name))
        return release

    def listDraftRelease(self, app_name, coll_id=None, revision=None, limit=Constant.DEFAULT_LIMIT, offset=0):
        """
        List the draft releases with an offset option to list only the older ones.

        By default only the first N releases are listed. Setting ``limit`` parameter to `0`
        removes this restriction.

        It's also possible to list one release within the Draft release by providing
        its specific revision.

        :param app_name: Name of the application
        :param coll_id: Collection ID
        :param revision: Revision of the release
        :param limit: Limit of the number of draft releases listed (see :const:`Constant.DEFAULT_LIMIT`)
        :param offset: offset to list only older revisions
        :return: The list of draft release
        """
        app = self._getApp(app_name=app_name, coll_id=coll_id)
        return self.get(
            '/app/%s/draft' % app['_id'],
            parameters={'revision': revision, 'limit': limit, 'offset': offset}
        )

    def deleteDraftRelease(self, app_name, revision, coll_id=None):
        """
        Delete a specific revision within the Draft release.

        :param app_name: Name of the application
        :param revision: Revision of the release
        :param coll_id: Collection ID
        :return: The deleted release
        """
        app = self._getApp(app_name=app_name, coll_id=coll_id)
        release = self.listDraftRelease(app_name, revision=revision)
        if not release:
            raise SlicerPackageManagerError(
                'The release with the revision "%s" doesn\'t exist.' % revision)
        self.delete('/app/%s/release/%s' % (app['_id'], release[0]['name']))
        return release[0]

    def uploadExtension(self, filepath, app_name, ext_os, arch, name, repo_type, repo_url,
                        revision, app_revision, desc='', packagetype='', icon_url='',
                        category=None, homepage='', screenshots=None, contributors=None,
                        dependency=None, coll_id=None, force=False):
        """
        Upload an extension by providing a path to the file. It can also be used to update an
        existing one, in this case the upload is done only if the extension has a different
        revision than the old one.

        :param filepath: The path to the file
        :param app_name: The name of the application
        :param ext_os: The target operating system of the package
        :param arch: The os chip architecture
        :param name: The baseName of the extension
        :param repo_type: Type of the repository
        :param repo_url: Url of the repository
        :param revision: The revision of the extension
        :param app_revision: The revision of the application supported by the extension
        :param desc: The description of the extension
        :param packagetype: Type of the package
        :param icon_url: Url of the extension's logo
        :param category: Category of the extension
        :param homepage: Url of the extension's homepage
        :param screenshots: Space-separate list of URLs of screenshots for the extension.
        :param contributors: List of contributors of the extension.
        :param dependency: List of the required extensions to use this one.
        :param coll_id: Collection ID
        :param force: To force update the binary file
        :return: The uploaded extension
        """
        def _displayProgress(*args, **kwargs):
            pass

        app = self._getApp(app_name=app_name, coll_id=coll_id)
        # Get potential existing extension
        extensions = self.listExtension(
            app_name,
            name=name,
            ext_os=ext_os,
            arch=arch,
            app_revision=app_revision)
        if not extensions:
            # Create the extension into Girder hierarchy
            extension = self.post('/app/%s/extension' % app['_id'], parameters={
                'os': ext_os,
                'arch': arch,
                'baseName': name,
                'repository_type': repo_type,
                'repository_url': repo_url,
                'revision': revision,
                'app_revision': app_revision,
                'description': desc,
                'packagetype': packagetype,
                'icon_url': icon_url,
                'category': category,
                'homepage': homepage,
                'screenshots': screenshots,
                'contributors': contributors,
                'dependency': dependency,
            })

            # Upload the extension
            self.uploadFileToItem(
                extension['_id'],
                filepath,
                reference='',
                mimeType='application/octet-stream',
                progressCallback=_displayProgress)
        else:
            extension = extensions[0]
            # Revision different or force upload
            if revision != extension['meta']['revision'] or force:
                files = list(self.listFile(extension['_id']))
                if files:
                    oldFile = files[0]
                    filename = 'new_file'
                else:
                    filename = None

                # Upload the extension
                newFile = self.uploadFileToItem(
                    extension['_id'],
                    filepath,
                    reference='',
                    filename=filename,
                    mimeType='application/octet-stream',
                    progressCallback=_displayProgress)

                # Update the extension into Girder hierarchy
                extension = self.post('/app/%s/extension' % app['_id'], parameters={
                    'os': ext_os,
                    'arch': arch,
                    'baseName': name,
                    'repository_type': repo_type,
                    'repository_url': repo_url,
                    'revision': revision,
                    'app_revision': app_revision,
                    'description': desc,
                    'packagetype': packagetype,
                    'icon_url': icon_url,
                    'category': category,
                    'homepage': homepage,
                    'screenshots': screenshots,
                    'contributors': contributors,
                    'dependency': dependency,
                })

                files = list(self.listFile(extension['_id']))
                if len(files) == 2:
                    # Remove the oldFIle
                    self.delete('/file/%s' % oldFile['_id'])
                    # Change the name
                    self.put('/file/%s' % newFile['_id'], parameters={
                        'name': os.path.basename(filepath)
                    })
                    return Constant.EXTENSION_NOW_UP_TO_DATE
            else:
                return Constant.EXTENSION_AREADY_UP_TO_DATE

        return extension

    def downloadExtension(self, app_name, id_or_name, coll_id=None,
                          dir_path=Constant.CURRENT_FOLDER):
        """
        Download an extension by ID and store it in the given option ``dir_path``.
        When we use the extension id in ``id_or_name``, the parameter ``app_name`` is ignored.

        :param app_name: Name of the application
        :param id_or_name: ID or name of the extension
        :param coll_id: Collection ID
        :param dir_path: Path of the directory where the extension has to be downloaded
        :return: The downloaded extension
        """
        return self._downloadPackage('extension', app_name=app_name, id_or_name=id_or_name,
                                     dir_path=dir_path, coll_id=coll_id)

    def listExtension(self, app_name, coll_id=None, name=None, ext_os=None, arch=None,
                      app_revision=None, release=Constant.DRAFT_RELEASE_NAME,
                      limit=Constant.DEFAULT_LIMIT, all=False):
        """
        List the extensions of a specific application ``app_name``.

        By default only the first N extensions within the ``draft`` release are listed. Setting ``limit``
        parameter to `0` removes this restriction.

        Specifying optional parameters like `ext_os` or `arch` allows to return the
        corresponding subset.

        Passing ``all=True`` option allow to list all the extensions from all the
        releases of an application.

        :param app_name: Name of the application
        :param coll_id: Collection ID
        :param name: Base name of the extension
        :param ext_os: The target operating system of the package
        :param arch: The os chip architecture
        :param app_revision: Revision of the application
        :param release: Name of the release
        :param limit: Limit of the number of extensions listed (see :const:`Constant.DEFAULT_LIMIT`)
        :param all: Boolean that allow to list extensions from all the release
        :return: A list of extensions filtered by optional parameters
        """
        app = self._getApp(app_name=app_name, coll_id=coll_id)

        if all:
            release_id = None
            limit = 0
        else:
            release_folder = self.listRelease(app_name, release)
            if release_folder:
                release_id = release_folder['_id']
            else:
                raise SlicerPackageManagerError(
                    'The release "%s" doesn\'t exist.' % release)

        extensions = self.get('/app/%s/extension' % app['_id'], parameters={
            'os': ext_os,
            'arch': arch,
            'baseName': name,
            'app_revision': app_revision,
            'release_id': release_id,
            'limit': limit,
            'sort': 'created',
            'sortDir': -1
        })
        return extensions

    def deleteExtension(self, app_name, id_or_name, coll_id=None):
        """
        Delete an extension within an application.

        :param app_name: Name of the application
        :param id_or_name: Extension ID or name
        :param coll_id: Collection ID
        :return: The deleted extension
        """
        return self._deletePackage('extension', app_name, id_or_name, coll_id)

    def uploadApplicationPackage(self, filepath, app_name, pkg_os, arch, name, repo_type,
                                 repo_url, revision, coll_id=None, desc='', pre_release=False):
        """
        Upload an application package by providing a path to the file.
        It can also be used to update an existing one.

        :param filepath: The path to the file
        :param app_name: The name of the application
        :param pkg_os: The target operating system of the package
        :param arch: The os chip architecture
        :param name: The baseName of the package
        :param repo_type: Type of the repository
        :param repo_url: Url of the repository
        :param revision: The revision of the application
        :param coll_id: Collection ID
        :param desc: The description of the application package
        :param pre_release: Boolean to specify if the package is ready to be distributed
        :return: The uploaded application package
        """
        def _displayProgress(*args, **kwargs):
            pass

        app = self._getApp(app_name=app_name, coll_id=coll_id)
        # Get potential existing package
        package = self.listApplicationPackage(
            app_name,
            name=name,
            pkg_os=pkg_os,
            arch=arch,
            revision=revision)
        if not package:
            # Create the package into Girder hierarchy
            package = self.post('/app/%s/package' % app['_id'], parameters={
                'os': pkg_os,
                'arch': arch,
                'baseName': name,
                'repository_type': repo_type,
                'repository_url': repo_url,
                'revision': revision,
                'description': desc,
                'pre_release': pre_release
            })

            # Upload the package
            self.uploadFileToItem(
                package['_id'],
                filepath,
                reference='',
                mimeType='application/octet-stream',
                progressCallback=_displayProgress)
        else:
            package = package[0]
            # Revision different or force upload
            files = list(self.listFile(package['_id']))
            if files:
                oldFile = files[0]
                filename = 'new_file'
            else:
                filename = None

            # Upload the package
            newFile = self.uploadFileToItem(
                package['_id'],
                filepath,
                reference='',
                filename=filename,
                mimeType='application/octet-stream',
                progressCallback=_displayProgress)

            # Update the package into Girder hierarchy
            package = self.post('/app/%s/package' % app['_id'], parameters={
                'os': pkg_os,
                'arch': arch,
                'baseName': name,
                'repository_type': repo_type,
                'repository_url': repo_url,
                'revision': revision,
                'description': desc
            })

            files = list(self.listFile(package['_id']))
            if len(files) == 2:
                # Remove the oldFIle
                self.delete('/file/%s' % oldFile['_id'])
                # Change the name
                self.put('/file/%s' % newFile['_id'], parameters={
                    'name': os.path.basename(filepath)
                })
                return Constant.PACKAGE_NOW_UP_TO_DATE
        return package

    def downloadApplicationPackage(self, app_name, id_or_name, coll_id=None,
                                   dir_path=Constant.CURRENT_FOLDER):
        """
        Download an application package by ID and store it in the given option ``dir_path``.
        When we use the package id in ``id_or_name``, the parameter ``app_name`` is ignored.

        :param app_name: Name of the application
        :param id_or_name: ID or name of the package
        :param coll_id: Collection ID
        :param dir_path: Path of the directory where the application package has to be downloaded
        :return: The downloaded package
        """
        return self._downloadPackage('package', app_name=app_name, id_or_name=id_or_name,
                                     dir_path=dir_path, coll_id=coll_id)

    def listApplicationPackage(self, app_name, coll_id=None, name=None, pkg_os=None, arch=None,
                               revision=None, release=None, limit=Constant.DEFAULT_LIMIT):
        """
        List the application packages filtered by some optional parameters (os, arch, ...).

        By default only the first N application packages are listed. Setting the ``limit`` parameter
        to `0` removes this restriction.

        It's also possible to specify the ``--release`` option to list all the package from a
        specific release.

        :param app_name: Name of the application
        :param coll_id: Collection ID
        :param name: Base name of the application package
        :param pkg_os: The target operating system of the package
        :param arch: The os chip architecture
        :param revision: Revision of the application
        :param release: Name or ID of the release
        :param limit: Limit of the number of applications listed (see :const:`Constant.DEFAULT_LIMIT`)
        :return: A list of application package filtered by optional parameters
        """
        app = self._getApp(app_name=app_name, coll_id=coll_id)
        if release and not ObjectId.is_valid(release):
            if not self.listRelease(app_name, release):
                raise SlicerPackageManagerError(
                    'The release "%s" doesn\'t exist.' % release)

        pkg = self.get('/app/%s/package' % app['_id'], parameters={
            'os': pkg_os,
            'arch': arch,
            'baseName': name,
            'revision': revision,
            'release_id_or_name': release,
            'limit': limit,
            'sort': 'created',
            'sortDir': -1
        })
        return pkg

    def deleteApplicationPackage(self, app_name, id_or_name, coll_id=None):
        """
        Delete an application package within an application.

        :param app_name: Name of the application
        :param id_or_name: Package ID or name
        :param coll_id: Collection ID
        :return: The deleted application package
        """
        return self._deletePackage(
            'package',
            app_name=app_name,
            id_or_name=id_or_name,
            coll_id=coll_id)

    # ---------------- UTILITIES ---------------- #

    def _getApp(self, app_name, coll_id=None):
        """
        Private method to get a single application by Name.

        :param app_name: Name of the application
        :param coll_id: ID of the collection that contains the application
        :return: A single application
        """
        apps = self.listApp(name=app_name, coll_id=coll_id)
        if not apps:
            raise SlicerPackageManagerError(
                'The Application "%s" doesn\'t exist.' % app_name)
        return apps[0]

    def _downloadPackage(self, package_type, app_name, id_or_name, dir_path, coll_id):
        app = self._getApp(app_name=app_name, coll_id=coll_id)

        if ObjectId.is_valid(id_or_name):
            pkg = self.get('/resource/%s' % id_or_name, parameters={'type': 'item'})
        else:
            pkg = self.get(
                '/app/%s/%s' % (app['_id'], package_type),
                parameters={'%s_name' % package_type: id_or_name})
            if pkg:
                pkg = pkg[0]
        if not pkg:
            raise SlicerPackageManagerError(
                'The %s "%s" doesn\'t exist.' % (package_type, id_or_name))
        files = self.get('/item/%s/files' % pkg['_id'])
        if not files:
            raise SlicerPackageManagerError(
                'The %s "%s" doesn\'t contain any file.' % (package_type, id_or_name))
        file = files[0]
        self.downloadFile(
            file['_id'],
            os.path.join(dir_path, '%s.%s' % (pkg['name'], file['name'].split('.')[1])))
        return pkg

    def _deletePackage(self, package_type, app_name, id_or_name, coll_id):
        app = self._getApp(app_name=app_name, coll_id=coll_id)

        if ObjectId.is_valid(id_or_name):
            pkg = self.get(
                '/app/%s/%s' % (app['_id'], package_type),
                parameters={'%s_id' % package_type: id_or_name})
        else:
            pkg = self.get(
                '/app/%s/%s' % (app['_id'], package_type),
                parameters={'%s_name' % package_type: id_or_name})
        if not pkg:
            raise SlicerPackageManagerError(
                'The %s "%s" doesn\'t exist.' % (package_type, id_or_name))
        pkg = pkg[0]
        self.delete('/app/%s/%s/%s' % (app['_id'], package_type, pkg['_id']))
        return pkg



# -*- coding: utf-8 -*-

import re


def extractID(file):
    with open(file, 'r') as f:
        res = re.search(r'\([a-z0-9]*\)', f.read())
        if res:
            id = res.group(0)
            return id[1:-1]


if __name__ == '__main__':
    from sys import argv
    print(extractID(argv[1]))



import os

import pytest

from bson.objectid import ObjectId

from girder.models.collection import Collection
from girder.models.folder import Folder
from girder.models.file import File
from girder.models.user import User

from pytest_girder.assertions import assertStatusOk
from pytest_girder.utils import getResponseBody

from slicer_package_manager import constants

from . import computeFileChecksum, computeContentChecksum, extensions, FIXTURE_DIR, packages, expectedDownloadStats


def _initialize():
    _user = User().createUser('usr0', 'passwd', 'tst', 'usr', 'u@u.com')
    _collection = Collection().createCollection(
        'testCollection',
        creator=_user,
        description='Contain applications')
    _packages = Folder().createFolder(
        parent=_collection,
        name=constants.TOP_LEVEL_FOLDER_NAME,
        parentType='Collection',
        public=True,
        creator=_user)
    _app = Folder().createFolder(
        parent=_packages,
        name='application',
        description='app description',
        parentType='Folder',
        public=True,
        creator=_user)
    _app = Folder().setMetadata(
        _app,
        {
            'applicationPackageNameTemplate': constants.APPLICATION_PACKAGE_TEMPLATE_NAME,
            'extensionPackageNameTemplate': constants.EXTENSION_PACKAGE_TEMPLATE_NAME
        }
    )
    _draftRelease = Folder().createFolder(
        parent=_app,
        name=constants.DRAFT_RELEASE_NAME,
        description='Uploaded each night, always up-to-date',
        parentType='Folder',
        public=True,
        creator=_user)
    _draftRevision = Folder().createFolder(
        parent=_draftRelease,
        name='0000',
        parentType='Folder',
        public=True,
        creator=_user)
    _draftRevision = Folder().setMetadata(
        _draftRevision,
        {'revision': '0000'}
    )
    _release = Folder().createFolder(
        parent=_app,
        name='release1',
        description='release description',
        parentType='Folder',
        public=True,
        creator=_user)

    _release = Folder().setMetadata(_release, {'revision': '0005'})
    _extensions = extensions
    _extensions['extension1']['name'] = constants.EXTENSION_PACKAGE_TEMPLATE_NAME.format(
        **_extensions['extension1']['meta'])
    _extensions['extension2']['name'] = constants.EXTENSION_PACKAGE_TEMPLATE_NAME.format(
        **_extensions['extension2']['meta'])
    _extensions['extension3']['name'] = constants.EXTENSION_PACKAGE_TEMPLATE_NAME.format(
        **_extensions['extension3']['meta'])
    _extensions['extension4']['name'] = constants.EXTENSION_PACKAGE_TEMPLATE_NAME.format(
        **_extensions['extension4']['meta'])
    _extensions['extension5']['name'] = constants.EXTENSION_PACKAGE_TEMPLATE_NAME.format(
        **_extensions['extension5']['meta'])
    _packages = packages
    _packages['package1']['name'] = constants.APPLICATION_PACKAGE_TEMPLATE_NAME.format(
        **_packages['package1']['meta'])
    _packages['package2']['name'] = constants.APPLICATION_PACKAGE_TEMPLATE_NAME.format(
        **_packages['package2']['meta'])
    _packages['package3']['name'] = constants.APPLICATION_PACKAGE_TEMPLATE_NAME.format(
        **_packages['package3']['meta'])

    return _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages


@pytest.mark.plugin('slicer_package_manager')
def testInitApp(server):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    _createApplicationCheck(
        server,
        'App_test',
        'Application without specifying any collection',
        collDescription='Automatic creation of the collection Applications',
        _user=_user
    )
    # Without any collection this should load the 'Applications' collection
    _createApplicationCheck(
        server,
        'App_test1',
        'Application without specifying any collection',
        _user=_user
    )
    # With a collection ID this should load the collection
    _createApplicationCheck(
        server,
        'App_test2',
        'Application with a collection ID',
        collId=_collection['_id'],
        _user=_user
    )
    # With a collection name that match an existing collection name
    _createApplicationCheck(
        server,
        'App_test3',
        'Application with a collection name',
        collName=_collection['name'],
        _user=_user
    )
    # With a collection name that does not exist yet
    _createApplicationCheck(
        server,
        'App_test4',
        'Application with a collection name',
        collName='Collection_for_app',
        _user=_user
    )


@pytest.mark.plugin('slicer_package_manager')
def testListApp(server):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    # Create applications in the 'Applications' collection
    app1 = _createApplicationCheck(
        server,
        'App_test1',
        'Application without specifying any collection',
        collDescription='Automatic creation of the collection Applications',
        _user=_user
    )
    app2 = _createApplicationCheck(
        server,
        'App_test2',
        'Application without specifying any collection',
        collDescription='Automatic creation of the collection Applications',
        _user=_user
    )
    # Get application with application ID
    resp = server.request(
        path='/app',
        method='GET',
        user=_user,
        params={'app_id': _app['_id']}
    )
    assertStatusOk(resp)
    assert ObjectId(resp.json['_id']) == _app['_id']
    assert resp.json['name'] == _app['name']
    assert resp.json['description'] == _app['description']
    assert resp.json['meta']['applicationPackageNameTemplate'] == constants.APPLICATION_PACKAGE_TEMPLATE_NAME
    assert resp.json['meta']['extensionPackageNameTemplate'] == constants.EXTENSION_PACKAGE_TEMPLATE_NAME
    # List all applications from 'Applications' collection (Default)
    resp = server.request(
        path='/app',
        method='GET',
        user=_user
    )
    assertStatusOk(resp)
    assert len(resp.json) == 2
    assert resp.json[0]['_id'] == app1['_id']
    assert resp.json[1]['_id'] == app2['_id']
    # Get application with exact name
    resp = server.request(
        path='/app',
        method='GET',
        user=_user,
        params={
            'collection_id': _collection['_id'],
            'name': _app['name']}
    )
    assertStatusOk(resp)
    assert ObjectId(resp.json[0]['_id']) == _app['_id']
    assert resp.json[0]['name'] == _app['name']
    assert resp.json[0]['description'] == _app['description']
    assert resp.json[0]['meta']['applicationPackageNameTemplate'] == constants.APPLICATION_PACKAGE_TEMPLATE_NAME
    assert resp.json[0]['meta']['extensionPackageNameTemplate'] == constants.EXTENSION_PACKAGE_TEMPLATE_NAME


@pytest.mark.plugin('slicer_package_manager')
def testDeleteApp(server):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    # Create the application without any collection,
    # this should create the new 'Applications' collection
    app = _createApplicationCheck(
        server,
        'App_test',
        'Application without specifying any collection',
        collDescription='Automatic creation of the collection Applications',
        _user=_user
    )
    # Get the new application by ID
    resp = server.request(
        path='/app',
        method='GET',
        user=_user,
        params={'app_id': app['_id']}
    )
    assertStatusOk(resp)
    assert resp.json['_id'] == app['_id']
    assert resp.json['name'] == app['name']
    # Delete the application
    resp = server.request(
        path='/app/%s' % app['_id'],
        method='DELETE',
        user=_user
    )
    assertStatusOk(resp)
    assert resp.json['_id'] == app['_id']
    # Try to get the application should failed
    resp = server.request(
        path='/app',
        method='GET',
        user=_user,
        params={'app_id': app['_id']}
    )
    assertStatusOk(resp)
    assert resp.json is None


@pytest.mark.plugin('slicer_package_manager')
def testNewRelease(server):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    _createReleaseCheck(
        server,
        name='V3.2.1',
        app_id=_app['_id'],
        app_revision='001',
        desc='This is a new release',
        _user=_user
    )


@pytest.mark.plugin('slicer_package_manager')
def testGetRelease(server):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    release1 = _createReleaseCheck(
        server,
        name='V3.2.1',
        app_id=_app['_id'],
        app_revision='002',
        desc='This is a new release',
        _user=_user
    )

    resp = server.request(
        path='/app/%s/release' % _app['_id'],
        method='GET',
        user=_user
    )
    # Check if it has return all the stable releases
    assertStatusOk(resp)
    assert len(resp.json) == 2
    assert resp.json[0]['_id'] == release1['_id']
    assert ObjectId(resp.json[1]['_id']) == _release['_id']

    # Get one release by ID
    resp = server.request(
        path='/app/%s/release' % _app['_id'],
        params={'release_id_or_name': _release['_id']},
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert resp.json['name'] == _release['name']
    assert resp.json['description'] == _release['description']
    assert ObjectId(resp.json['_id']) == _release['_id']

    # Get one release by name
    resp = server.request(
        path='/app/%s/release' % _app['_id'],
        params={'release_id_or_name': _release['name']},
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert resp.json['name'] == _release['name']
    assert resp.json['description'] == _release['description']
    assert ObjectId(resp.json['_id']) == _release['_id']

    resp = server.request(
        path='/app/%s/release' % _app['_id'],
        params={'release_id_or_name': constants.DRAFT_RELEASE_NAME},
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert resp.json['name'] == constants.DRAFT_RELEASE_NAME
    assert ObjectId(resp.json['_id']) == _draftRelease['_id']


@pytest.mark.plugin('slicer_package_manager')
def testGetAllDraftRelease(server):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    resp = server.request(
        path='/app/%s/draft' % _app['_id'],
        method='GET',
        user=_user
    )
    # Check if it has return all the revision from the default release
    assertStatusOk(resp)
    assert len(resp.json) == 1
    assert ObjectId(resp.json[0]['_id']) == _draftRevision['_id']
    assert resp.json[0]['meta']['revision'] == _draftRevision['meta']['revision']
    # With the parameter 'revision'
    resp = server.request(
        path='/app/%s/draft' % _app['_id'],
        method='GET',
        user=_user,
        params={'revision': _draftRevision['meta']['revision']}
    )
    # Check if it has return the good revision from the draft folder
    assertStatusOk(resp)
    assert len(resp.json) == 1
    assert ObjectId(resp.json[0]['_id']) == _draftRevision['_id']
    assert resp.json[0]['meta']['revision'] == _draftRevision['meta']['revision']
    # With the parameter 'revision' set to a wrong value
    resp = server.request(
        path='/app/%s/draft' % _app['_id'],
        method='GET',
        user=_user,
        params={'revision': 'wrongRev'}
    )
    # Check if it has return the good revision from the draft folder
    assertStatusOk(resp)
    assert len(resp.json) == 0


@pytest.mark.plugin('slicer_package_manager')
def testDeleteReleaseByID(server):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    _deleteRelease(server, '_id', _user=_user, _app=_app)


@pytest.mark.plugin('slicer_package_manager')
def testDeleteReleaseByName(server):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    _deleteRelease(server, 'name', _user=_user, _app=_app)


@pytest.mark.external_data(
    os.path.join(FIXTURE_DIR, 'extension3.tar.gz'),
    os.path.join(FIXTURE_DIR, 'extension4.tar.gz'),
)
@pytest.mark.plugin('slicer_package_manager')
def testDeleteRevisionRelease(server, fsAssetstore, external_data):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    # Create extensions in the "draft" release
    extension1 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension3']['meta'],
        external_data.join('extension3.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension1['name'] == _extensions['extension3']['name']
    extension2 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension4']['meta'],
        external_data.join('extension4.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension2['name'] == _extensions['extension4']['name']

    resp = server.request(
        path='/app/%s/draft' % _app['_id'],
        method='GET',
        user=_user
    )
    # Check if it has return all the revision from the default release
    assertStatusOk(resp)
    assert len(resp.json) == 2

    # Delete by Name the revision release '0001' in the "draft" release
    resp = server.request(
        path='/app/%s/release/%s' %
             (_app['_id'], _extensions['extension4']['meta']['app_revision']),
        method='DELETE',
        user=_user
    )
    assertStatusOk(resp)

    resp = server.request(
        path='/app/%s/draft' % _app['_id'],
        method='GET',
        user=_user
    )
    # Check if it has return all the revision from the default release
    assertStatusOk(resp)
    assert len(resp.json) == 1


@pytest.mark.external_data(
    os.path.join(FIXTURE_DIR, 'extension1.tar.gz'),
    os.path.join(FIXTURE_DIR, 'extension2.tar.gz'),
    os.path.join(FIXTURE_DIR, 'extension3.tar.gz')
)
@pytest.mark.plugin('slicer_package_manager')
def testUploadAndDownloadExtension(server, fsAssetstore, external_data):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    # Create a new extension in the release "_release"
    extension1 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension1']['meta'],
        external_data.join('extension1.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension1['name'] == _extensions['extension1']['name']
    extensions_folder = Folder().load(extension1['folderId'], user=_user)
    assert ObjectId(extensions_folder['parentId']) == _release['_id']
    # Create an other extension in the "draft" release
    extension2 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension2']['meta'],
        external_data.join('extension2.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension2['name'] == _extensions['extension2']['name']
    # Create a third extension in the "draft" release
    extension3 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension3']['meta'],
        external_data.join('extension3.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension3['name'] == _extensions['extension3']['name']
    # Try to create the same extension should just get the same one
    extension3 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension3']['meta'],
        external_data.join('extension3.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension3['name'] == _extensions['extension3']['name']


@pytest.mark.external_data(
    os.path.join(FIXTURE_DIR, 'extension2.tar.gz')
)
@pytest.mark.plugin('slicer_package_manager')
def testUpdateExtensions(server, fsAssetstore, external_data):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    extension = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension2']['meta'],
        external_data.join('extension2.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension['name'] == _extensions['extension2']['name']
    # Update the same extension
    newParams = _extensions['extension2']['meta'].copy()
    newParams.update({
        'revision': '0000',
        'repository_type': 'gitlab',
        'packagetype': 'zip',
        'codebase': 'SL434334',
        'description': 'Extension for Slicer 4 new version 2'
    })
    updatedExtension = _createOrUpdatePackage(server, 'extension', newParams, _user=_user, _app=_app)
    # Check the same extension has different metadata
    assert updatedExtension['_id'] == extension['_id']
    assert updatedExtension['name'] == constants.EXTENSION_PACKAGE_TEMPLATE_NAME.format(**newParams)
    assert updatedExtension['meta'] != extension['meta']


@pytest.mark.plugin('slicer_package_manager')
def testGetExtensions(server):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    # Create a new extension in the release "_release"
    extension1 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension1']['meta'],
        _user=_user,
        _app=_app
    )
    assert extension1['name'] == _extensions['extension1']['name']
    extensions_folder = Folder().load(extension1['folderId'], user=_user)
    assert ObjectId(extensions_folder['parentId']) == _release['_id']
    # Create other extensions in the "draft" release
    extension2 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension2']['meta'],
        _user=_user,
        _app=_app
    )
    assert extension2['name'] == _extensions['extension2']['name']
    extension3 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension3']['meta'],
        _user=_user,
        _app=_app
    )
    assert extension3['name'] == _extensions['extension3']['name']
    # Get all the extension of the application
    resp = server.request(
        path='/app/%s/extension' % _app['_id'],
        method='GET',
        user=_user
    )
    assertStatusOk(resp)
    assert len(resp.json) == 3
    # Get all the extension of the application for Linux
    resp = server.request(
        path='/app/%s/extension' % _app['_id'],
        method='GET',
        user=_user,
        params={
            'os': 'linux'
        }
    )
    assertStatusOk(resp)
    assert len(resp.json) == 2
    for ext in resp.json:
        assert ext['meta']['os'] == 'linux'
    # Get all the extension of the application for Linux and amd64 architecture
    resp = server.request(
        path='/app/%s/extension' % _app['_id'],
        method='GET',
        user=_user,
        params={
            'os': 'linux',
            'arch': 'amd64'
        }
    )
    assertStatusOk(resp)
    assert len(resp.json) == 1
    for ext in resp.json:
        assert ext['meta']['os'] == 'linux'
        assert ext['meta']['arch'] == 'amd64'
    # Get all the extension of the application which are in "release1"
    resp = server.request(
        path='/app/%s/extension' % _app['_id'],
        method='GET',
        user=_user,
        params={
            'release_id': _release['_id']
        }
    )
    assertStatusOk(resp)
    assert len(resp.json) == 1
    # Get all the extension of the application which are in the draft release
    draftRelease = list(Folder().childFolders(
        _app,
        'Folder',
        user=_user,
        filters={'name': constants.DRAFT_RELEASE_NAME}))
    resp = server.request(
        path='/app/%s/extension' % _app['_id'],
        method='GET',
        user=_user,
        params={
            'release_id': draftRelease[0]['_id']
        }
    )
    assertStatusOk(resp)
    assert len(resp.json) == 2
    # Get a specific extension by name
    resp = server.request(
        path='/app/%s/extension' % _app['_id'],
        params={'extension_name': extension3['name']},
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert resp.json[0]['_id'] == extension3['_id']
    assert resp.json[0]['name'] == extension3['name']
    # Get a specific extension with wrong name
    resp = server.request(
        path='/app/%s/extension' % _app['_id'],
        params={'extension_name': 'wrong_name'},
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert len(resp.json) == 0
    assert resp.json == []


@pytest.mark.plugin('slicer_package_manager')
def testDeleteExtensionPackages(server):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    # Create a new extension in the release "_release"
    extension = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension1']['meta'],
        _user=_user,
        _app=_app
    )
    assert extension['name'] == _extensions['extension1']['name']
    extensions_folder = Folder().load(extension['folderId'], user=_user)
    assert ObjectId(extensions_folder['parentId']) == _release['_id']
    # Get, delete, and try to re-get the package
    _deletePackages(server, 'extension', extension, _user=_user, _app=_app)


@pytest.mark.external_data(
    os.path.join(FIXTURE_DIR, 'pkg1.dmg'),
    os.path.join(FIXTURE_DIR, 'pkg2.exe'),
    os.path.join(FIXTURE_DIR, 'pkg3.tar.gz')
)
@pytest.mark.plugin('slicer_package_manager')
def testUploadAndDownloadPackages(server, fsAssetstore, external_data):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    # Create a new application package in the release "_release"
    package1 = _createOrUpdatePackage(
        server,
        'package',
        _packages['package1']['meta'],
        external_data.join('pkg1.dmg'),
        _user=_user,
        _app=_app
    )
    assert package1['name'] == _packages['package1']['name']
    assert ObjectId(package1['folderId']) == _release['_id']
    # Create an other application package in the "draft" release
    package2 = _createOrUpdatePackage(
        server,
        'package',
        _packages['package2']['meta'],
        external_data.join('pkg2.exe'),
        _user=_user,
        _app=_app
    )
    assert package2['name'] == _packages['package2']['name']
    # Create a third application package in the "draft" release
    package3 = _createOrUpdatePackage(
        server,
        'package',
        _packages['package3']['meta'],
        external_data.join('pkg3.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert package3['name'] == _packages['package3']['name']
    # Try to create the same application package should just get the same one
    package3 = _createOrUpdatePackage(
        server,
        'package',
        _packages['package3']['meta'],
        external_data.join('pkg3.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert package3['name'] == _packages['package3']['name']


@pytest.mark.external_data(
    os.path.join(FIXTURE_DIR, 'pkg2.exe'),
    )
@pytest.mark.plugin('slicer_package_manager')
def testUpdatePackages(server, fsAssetstore, external_data):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    package = _createOrUpdatePackage(
        server,
        'package',
        _packages['package2']['meta'],
        external_data.join('pkg2.exe'),
        _user=_user,
        _app=_app
    )
    assert package['name'] == _packages['package2']['name']
    # Update the same package
    newParams = _packages['package2']['meta'].copy()
    newParams.update({
        'repository_url': 'https://AnotherURL.com',
        'repository_type': 'gitlab',
    })
    updatedPackage = _createOrUpdatePackage(server, 'package', newParams, _user=_user, _app=_app)
    # Check the same package has different metadata
    assert updatedPackage['_id'] == package['_id']
    assert updatedPackage['name'] == constants.APPLICATION_PACKAGE_TEMPLATE_NAME.format(**newParams)
    assert updatedPackage['meta'] != package['meta']


@pytest.mark.external_data(
    os.path.join(FIXTURE_DIR, 'pkg1.dmg'),
    os.path.join(FIXTURE_DIR, 'pkg2.exe'),
    os.path.join(FIXTURE_DIR, 'pkg3.tar.gz'),
    )
@pytest.mark.plugin('slicer_package_manager')
def testGetPackages(server, fsAssetstore, external_data):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    # Create a new application package in the release "_release"
    package1 = _createOrUpdatePackage(
        server,
        'package',
        _packages['package1']['meta'],
        external_data.join('pkg1.dmg'),
        _user=_user,
        _app=_app
    )
    assert package1['name'] == _packages['package1']['name']
    assert ObjectId(package1['folderId']) == _release['_id']
    # Create an other application package in the "draft" release
    package2 = _createOrUpdatePackage(
        server,
        'package',
        _packages['package2']['meta'],
        external_data.join('pkg2.exe'),
        _user=_user,
        _app=_app
    )
    assert package2['name'] == _packages['package2']['name']
    # Create a third application package in the "draft" release
    package3 = _createOrUpdatePackage(
        server,
        'package',
        _packages['package3']['meta'],
        external_data.join('pkg3.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert package3['name'] == _packages['package3']['name']
    # Get all the package of the application
    resp = server.request(
        path='/app/%s/package' % _app['_id'],
        method='GET',
        user=_user
    )
    assertStatusOk(resp)
    assert len(resp.json) == 3
    # Get all the package of the application for Linux
    resp = server.request(
        path='/app/%s/package' % _app['_id'],
        method='GET',
        user=_user,
        params={
            'os': 'linux'
        }
    )
    assertStatusOk(resp)
    assert len(resp.json) == 1
    assert resp.json[0]['meta']['os'] == 'linux'
    # Get all the package of the application for Linux and amd64 architecture
    resp = server.request(
        path='/app/%s/package' % _app['_id'],
        method='GET',
        user=_user,
        params={
            'os': 'macosx',
            'arch': 'i386'
        }
    )
    assertStatusOk(resp)
    assert len(resp.json) == 0
    # Get all the package of the application which are in "release1"
    resp = server.request(
        path='/app/%s/package' % _app['_id'],
        method='GET',
        user=_user,
        params={
            'release_id_or_name': _release['_id']
        }
    )
    assertStatusOk(resp)
    assert len(resp.json) == 1
    # Get all the package of the application which are in the draft release
    draftRelease = list(Folder().childFolders(
        _app,
        'Folder',
        user=_user,
        filters={'name': constants.DRAFT_RELEASE_NAME}))
    resp = server.request(
        path='/app/%s/package' % _app['_id'],
        method='GET',
        user=_user,
        params={
            'release_id_or_name': draftRelease[0]['_id']
        }
    )
    assertStatusOk(resp)
    assert len(resp.json) == 2
    # Get a specific package by name
    resp = server.request(
        path='/app/%s/package' % _app['_id'],
        params={'package_name': package3['name']},
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert resp.json[0]['_id'] == package3['_id']
    assert resp.json[0]['name'] == package3['name']
    # Get a specific extension with wrong name
    resp = server.request(
        path='/app/%s/package' % _app['_id'],
        params={'package_name': 'wrong_name'},
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert len(resp.json) == 0
    assert resp.json == []


@pytest.mark.plugin('slicer_package_manager')
def testDeleteApplicationPackages(server):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    # Create a new application package in the release "_release"
    package = _createOrUpdatePackage(
        server,
        'package',
        _packages['package1']['meta'],
        _user=_user,
        _app=_app
    )
    assert package['name'] == _packages['package1']['name']
    assert ObjectId(package['folderId']) == _release['_id']
    # Get, delete, and try to re-get the package
    _deletePackages(server, 'package', package, _user=_user, _app=_app)


@pytest.mark.external_data(
    os.path.join(FIXTURE_DIR, 'extension1.tar.gz'),
    os.path.join(FIXTURE_DIR, 'extension2.tar.gz'),
    os.path.join(FIXTURE_DIR, 'extension3.tar.gz'),
    os.path.join(FIXTURE_DIR, 'extension4.tar.gz'),
    os.path.join(FIXTURE_DIR, 'extension5.tar.gz'),
    os.path.join(FIXTURE_DIR, 'pkg1.dmg'),
    os.path.join(FIXTURE_DIR, 'pkg2.exe'),
    os.path.join(FIXTURE_DIR, 'pkg3.tar.gz'),
    )
@pytest.mark.plugin('slicer_package_manager')
def testDownloadStats(server, fsAssetstore, external_data):
    _user, _collection, _app, _release, _draftRelease, _draftRevision, _extensions, _packages = _initialize()
    extension1 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension1']['meta'],
        external_data.join('extension1.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension1['name'] == _extensions['extension1']['name']
    extension2 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension2']['meta'],
        external_data.join('extension2.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension2['name'] == _extensions['extension2']['name']
    extension3 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension3']['meta'],
        external_data.join('extension3.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension3['name'] == _extensions['extension3']['name']
    extension4 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension4']['meta'],
        external_data.join('extension4.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension4['name'] == _extensions['extension4']['name']
    extension5 = _createOrUpdatePackage(
        server,
        'extension',
        _extensions['extension5']['meta'],
        external_data.join('extension5.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert extension5['name'] == _extensions['extension5']['name']
    package1 = _createOrUpdatePackage(
        server,
        'package',
        _packages['package1']['meta'],
        external_data.join('pkg1.dmg'),
        _user=_user,
        _app=_app
    )
    assert package1['name'] == _packages['package1']['name']
    package2 = _createOrUpdatePackage(
        server,
        'package',
        _packages['package2']['meta'],
        external_data.join('pkg2.exe'),
        _user=_user,
        _app=_app
    )
    assert package2['name'] == _packages['package2']['name']
    package3 = _createOrUpdatePackage(
        server,
        'package',
        _packages['package3']['meta'],
        external_data.join('pkg3.tar.gz'),
        _user=_user,
        _app=_app
    )
    assert package3['name'] == _packages['package3']['name']

    # Get the downloadStats
    expectedStats = expectedDownloadStats
    resp = server.request(
        path='/app/%s/downloadstats' % _app['_id'],
        method='GET',
        user=_user
    )
    assertStatusOk(resp)
    assert resp.json == expectedStats

    # Download multiple time an extension
    ext4_file = list(File().find({'itemId': ObjectId(extension4['_id'])}))
    ext5_file = list(File().find({'itemId': ObjectId(extension5['_id'])}))

    N = 5
    for _idx in range(N):
        _downloadFile(server, ext4_file[0]['_id'], _user=_user)
        _downloadFile(server, ext5_file[0]['_id'], _user=_user)

    expectedStats['0001']['extensions']['Ext3']['macosx'].update({
        'amd64': N + expectedStats['0001']['extensions']['Ext3']['macosx']['amd64'],
        'i386': N + expectedStats['0001']['extensions']['Ext3']['macosx']['i386']
    })
    resp = server.request(
        path='/app/%s/downloadstats' % _app['_id'],
        method='GET',
        user=_user
    )
    assertStatusOk(resp)
    assert resp.json == expectedStats

    # The download stats aren't affected by deletion of extension or revision
    # in the default release
    # Delete extensions
    resp = server.request(
        path='/app/%(app_id)s/extension/%(ext_id)s' % {
            'app_id': _app['_id'],
            'ext_id': extension4['_id']},
        method='DELETE',
        user=_user
    )
    assertStatusOk(resp)
    resp = server.request(
        path='/app/%(app_id)s/extension/%(ext_id)s' % {
            'app_id': _app['_id'],
            'ext_id': extension5['_id']},
        method='DELETE',
        user=_user
    )
    assertStatusOk(resp)
    # Delete the revision '0000' in the "draft" release
    resp = server.request(
        path='/app/%(app_id)s/release/%(release_id)s' % {
            'app_id': _app['_id'],
            'release_id': _draftRevision['_id']},
        method='DELETE',
        user=_user
    )
    assertStatusOk(resp)
    # Check if download Stats are the same
    resp = server.request(
        path='/app/%s/downloadstats' % _app['_id'],
        method='GET',
        user=_user
    )
    assertStatusOk(resp)
    assert resp.json == expectedStats


def _createApplicationCheck(server, appName, appDescription, collId=None,
                            collName=None, collDescription='', _user=None):
    params = {
        'name': appName,
        'app_description': appDescription,
        'collection_description': collDescription
    }
    if not (collId or collName):
        collName = 'Applications'
        params.update({'collection_name': collName})
    elif collId:
        if not ObjectId.is_valid(collId):
            raise Exception("The ObjectId %s is not valid" % collId)
        params.update({'collection_id': collId})
        collection = Collection().load(collId, user=_user)
        collName = collection['name']
    else:
        params.update({'collection_name': collName})

    resp = server.request(
        path='/app',
        method='POST',
        user=_user,
        params=params
    )
    # Check if it has created the application
    assertStatusOk(resp)
    assert resp.json['name'] == appName
    assert resp.json['description'] == appDescription
    assert resp.json['meta']['applicationPackageNameTemplate'] == constants.APPLICATION_PACKAGE_TEMPLATE_NAME
    assert resp.json['meta']['extensionPackageNameTemplate'] == constants.EXTENSION_PACKAGE_TEMPLATE_NAME
    # Check if it has created/load the collection
    topLevelFolder = Folder().load(resp.json['parentId'], user=_user)
    collection = Collection().load(topLevelFolder['parentId'], user=_user)
    assert collection['name'] == collName
    if collDescription:
        assert collection['description'] == collDescription
    # Check if it has created the "draft" release
    parent = Folder().load(resp.json['_id'], user=_user)
    draft = list(Folder().childFolders(parent, 'Folder', user=_user))
    assert len(draft) == 1
    assert draft[0]['name'] == constants.DRAFT_RELEASE_NAME
    assert draft[0]['description'] == 'Uploaded each night, always up-to-date'

    return resp.json


def _createReleaseCheck(server, name, app_id, app_revision, desc='', _user=None):
    resp = server.request(
        path='/app/%s/release' % app_id,
        method='POST',
        user=_user,
        params={
            'name': name,
            'app_revision': app_revision,
            'description': desc
        }
    )
    # Check if it has created the new release
    assertStatusOk(resp)
    assert resp.json['name'] == name
    assert resp.json['description'] == desc
    assert resp.json['meta']['revision'] == app_revision
    return resp.json


def _deleteRelease(server, identifier, _user=None, _app=None):
    # Create the release
    release = _createReleaseCheck(
        server,
        name='V3.2.1',
        app_id=_app['_id'],
        app_revision='001',
        desc='This is a new release',
        _user=_user
    )
    # Get the new release by name
    resp = server.request(
        path='/app/%s/release' % _app['_id'],
        params={'release_id_or_name': release[identifier]},
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert resp.json['_id'] == release['_id']
    assert resp.json['name'] == release['name']
    # Delete the release by name
    resp = server.request(
        path='/app/%(app_id)s/release/%(id_or_name)s' % {
            'app_id': _app['_id'],
            'id_or_name': release[identifier]},
        method='DELETE',
        user=_user
    )
    assertStatusOk(resp)
    assert resp.json['_id'] == release['_id']
    # Try to get the release should failed
    resp = server.request(
        path='/app/%s/release' % _app['_id'],
        params={'release_id_or_name': release[identifier]},
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert resp.json is None


# def _uploadExternalData(server, item, filePath, fileName=None, _user=None):
#     file = os.path.join(self.dataDir, filePath)
#     size = os.path.getsize(file)
#     offset = 0
#     if fileName is None:
#         fileName = filePath
#
#     # Initialize the upload
#     resp = server.request(
#         path='/file', method='POST', user=_user, params={
#             'parentType': 'item',
#             'parentId': item['_id'],
#             'name': fileName,
#             'size': size,
#             'mimeType': 'application/octet-stream'
#         })
#     assertStatusOk(resp)
#     uploadId = resp.json['_id']
#
#     # Upload content
#     with open(file, 'rb') as fp:
#         while True:
#             chunk = fp.read(min(self._MAX_CHUNK_SIZE, (size - offset)))
#             if not chunk:
#                 break
#             if isinstance(chunk, str):
#                 chunk = chunk.encode('utf8')
#             resp = server.request(
#                 path='/file/chunk',
#                 method='POST',
#                 user=_user,
#                 params={
#                     'offset': offset,
#                     'uploadId': uploadId
#                 },
#                 body=chunk,
#                 type='application/octet-stream'
#             )
#             assertStatusOk(resp)
#             offset += len(chunk)
#
#         uploadedFile = resp.json
#         assert 'itemId' in uploadedFile
#         assert uploadedFile['assetstoreId'] == str(self.assetstore['_id'])
#         assert uploadedFile['name'] == filePath
#         assert uploadedFile['size'] == size
#     return uploadedFile


def _downloadFile(server, id, _user=None):
    resp = server.request(
        path='/file/%s/download' % id,
        method='GET',
        user=_user,
        isJson=False)
    assertStatusOk(resp)
    return getResponseBody(resp, text=False)


def _createOrUpdatePackage(server, packageType, params, filePath=None, _user=None, _app=None):
    resp = server.request(
        path='/app/%s/%s' % (_app['_id'], packageType),
        method='POST',
        user=_user,
        params=params)
    assertStatusOk(resp)

    # Remove the app_id to check the metadata with these set during the creation
    metadata = resp.json['meta'].copy()
    metadata.pop('app_id')
    assert metadata == params

    if filePath:
        # Upload a package file
        item = resp.json
        with open(filePath, 'rb') as content:
            uploadedFile = server.uploadFile(
                filePath, content.read(), _user, item, parentType="item")

        # Download the package and compare checksum
        downloaded_checksum = computeContentChecksum('SHA512', _downloadFile(server, uploadedFile['_id'], _user=_user))
        local_checksum = computeFileChecksum('SHA512', filePath)
        assert downloaded_checksum == local_checksum
    return resp.json


def _deletePackages(server, packageType, pkg, _user=None, _app=None):
    # Get all the  application package, this should only be the new one
    resp = server.request(
        path='/app/%s/%s' % (_app['_id'], packageType),
        method='GET',
        user=_user
    )
    assertStatusOk(resp)
    assert len(resp.json) == 1
    assert resp.json[0]['_id'] == pkg['_id']
    # Delete the package
    resp = server.request(
        path='/app/%(app_id)s/%(type)s/%(pkg_id)s' % {
            'app_id': _app['_id'],
            'type': packageType,
            'pkg_id': pkg['_id']},
        method='DELETE',
        user=_user
    )
    assertStatusOk(resp)
    assert resp.json['_id'] == pkg['_id']
    # Try to get the package shouldn't success
    resp = server.request(
        path='/app/%s/%s' % (_app['_id'], packageType),
        method='GET',
        user=_user
    )
    assertStatusOk(resp)
    assert len(resp.json) == 0



# -*- coding: utf-8 -*-

from girder_client import GirderClient

if __name__ == "__main__":
    login = 'admin'
    password = 'adminadmin'

    gc = GirderClient(apiUrl='http://localhost:8080/api/v1')

    # Create an admin user & authenticate
    gc.createUser(
        login=login,
        email='admin@admin.com',
        firstName='admin',
        lastName='admin',
        password=password,
        admin=True
    )
    gc.authenticate(username=login, password=password)

    # Create an assetstore
    gc.post('assetstore', parameters={
        'name': 'TestAssetstore',
        'type': 0,
        'root': '/home/circleci/project/assetstore'
    })

    # Enable the 'slicer_package_manager' plugin
    gc.put('system/plugins', parameters={
        "plugins": '["slicer_package_manager"]'
    })

    # Restart the server
    gc.put('system/restart')



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

install_reqs = [
    'click>=6.7',
    'girder_client',
    'pytest-vcr',
    'tabulate'
]
with open('README.rst') as f:
    readme = f.read()

# perform the install
setup(
    name='slicer_package_manager_client',
    version='0.0.1',
    description='Python client for interacting with the Slicer package manager endpoint.',
    long_description=readme,
    author='Kitware, Inc.',
    author_email='kitware@kitware.com',
    url='https://slicer-package-manager.readthedocs.io/en/latest/commands_shell.html#slicer-package-manager-client',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3'
    ],
    packages=find_packages(exclude=('tests.*', 'tests')),
    install_requires=install_reqs,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'slicer_package_manager_client = slicer_package_manager_client.cli:main'
        ]
    }
)



# -*- coding: utf-8 -*-

import versioneer

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    'girder>=3.0.0a1'
]

setup(
    author='Jean-Christophe Fillion-Robin',
    author_email='slicer-packages-support@kitware.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    description='Manage Slicer application and extension packages.',
    install_requires=requirements,
    license='Apache Software License 2.0',
    long_description=readme,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    keywords='girder-plugin, slicer_package_manager',
    name='slicer_package_manager',
    packages=find_packages(exclude=['test', 'test.*']),
    url='https://github.com/girder/slicer_package_manager',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    zip_safe=False,
    entry_points={
        'girder.plugin': [
            'slicer_package_manager = slicer_package_manager:GirderPlugin'
        ]
    }
)



# -*- coding: utf-8 -*-

import datetime
import six

from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.model_base import ValidationException


class Package(Item):
    """
    The ``Package`` class derive from the ``Item`` model in Girder and it uses to embedded
    packages binary files.
    """

    def initialize(self):
        super(Package, self).initialize()
        # To be able to upload within a Package the name has to stay as 'item'.
        self.name = 'item'

    def get(self, release, limit=0, offset=0, sort=None, filters=None, **kwargs):
        """
        Get all the application packages available for a release.

        :param release: The release folder
        :return: Generator containing all the child items of the release.
        """
        return Folder().childItems(
            release,
            limit=limit,
            offset=offset,
            sort=sort,
            filters=filters,
            **kwargs)

    def createPackage(self, name, creator, folder, params, description):
        """
        Create and save in the DB a new application package.

        :param name: Name of the new package
        :param creator: The creator user
        :param folder: The release folder within the package will be created.
        :param params: All the metadata to set on the new application package
        :return: The new application package item
        """
        item = self.createItem(
            name,
            creator,
            folder,
            description
        )
        return self.setMetadata(item, params)

    def validate(self, doc):
        """
        Validate the package instance.

        :param doc: The package instance
        :return: The package instance once validated
        """
        # Call Item validate method
        doc = super(Package, self).validate(doc)

        if not isinstance(doc.get('created'), datetime.datetime):
            raise ValidationException(
                'Package field "created" must be a datetime.')

        # Validate the meta field
        if doc.get('meta'):
            base_params = {'app_id', 'os', 'arch', 'repository_type', 'repository_url',
                           'revision', 'baseName'}
            specs = []
            for meta in base_params:
                specs.append({
                    'name': meta,
                    'type': six.string_types,
                    'exception_msg': 'Package field "%s" must be a non-empty string.' % meta
                })
            for spec in specs:
                if doc['meta'].get(spec['name']) and not isinstance(
                   doc['meta'][spec['name']], spec['type']):
                    raise ValidationException(spec['exception_msg'])
            duplicateQuery = {
                'name': doc['name'],
                'os': doc['meta']['os'],
                'arch': doc['meta']['arch'],
                'revision': doc['meta']['revision']
            }
            if '_id' in doc:
                duplicateQuery['_id'] = {'$ne': doc['_id']}
            if self.findOne(duplicateQuery, fields=['_id']):
                raise ValidationException(
                    'A Package with this name and characteristics already exists.')
        return doc



# -*- coding: utf-8 -*-

import time
import os
import pytest

from slicer_package_manager_client import SlicerPackageClient, SlicerPackageManagerError


APPS = ['App', 'App1', 'App2']

RELEASES = [
    {
        'app_name': APPS[0],
        'name': 'Release',
        'revision': 'r000',
        'desc': 'random description 1',
    },
    {
        'app_name': APPS[0],
        'name': 'Release1',
        'revision': 'r001',
        'desc': 'random description 2',
    }
]

DRAFT_RELEASES = [
    {
        'revision': 'extR001'
    },
    {
        'revision': 'extR002'
    }
]

PACKAGES = [
    {
        'filepath': './file1.txt',
        'app_name': APPS[0],
        'os': 'macosx',
        'arch': 'amd64',
        'baseName': 'pkg1',
        'repo_type': 'git',
        'repo_url': 'git@github.com:pkg1.git',
        'revision': RELEASES[0]['revision']
    },
    {
        'filepath': './file2.txt',
        'app_name': APPS[0],
        'os': 'linux',
        'arch': 'amd64',
        'baseName': 'pkg2',
        'repo_type': 'git',
        'repo_url': 'git@github.com:pkg2.git',
        'revision': RELEASES[0]['revision']
    },
    {
        'filepath': './file3.txt',
        'app_name': APPS[0],
        'os': 'win',
        'arch': 'i386',
        'baseName': 'pkg3',
        'repo_type': 'git',
        'repo_url': 'git@github.com:pkg3.git',
        'revision': RELEASES[0]['revision']
    }
]

EXTENSIONS = [
    {
        'filepath': './file1.txt',
        'app_name': APPS[0],
        'os': 'macosx',
        'arch': 'amd64',
        'baseName': 'ext1',
        'repo_type': 'git',
        'repo_url': 'git@github.com:ext1.git',
        'app_revision': DRAFT_RELEASES[0]['revision'],
        'revision': 'r300'
    },
    {
        'filepath': './file2.txt',
        'app_name': APPS[0],
        'os': 'linux',
        'arch': 'amd64',
        'baseName': 'ext2',
        'repo_type': 'git',
        'repo_url': 'git@github.com:ext2.git',
        'app_revision': DRAFT_RELEASES[1]['revision'],
        'revision': 'r301'
    },
    {
        'filepath': './file3.txt',
        'app_name': APPS[0],
        'os': 'win',
        'arch': 'i386',
        'baseName': 'ext3',
        'repo_type': 'git',
        'repo_url': 'git@github.com:ext3.git',
        'app_revision': RELEASES[0]['revision'],
        'revision': 'r302'
    }
]


@pytest.mark.vcr()
@pytest.fixture
def spc(server):
    spc = SlicerPackageClient(apiUrl='http://localhost:8080/api/v1')
    spc.authenticate('admin', 'password')
    yield spc


@pytest.mark.vcr()
@pytest.fixture
def apps(server, spc):
    app1 = spc.createApp(name=APPS[0], desc='random description 1')
    app2 = spc.createApp(name=APPS[1], desc='random description 2')
    yield [app1, app2]


@pytest.mark.vcr()
@pytest.fixture
def releases(server, spc):

    def _create(release):
        return spc.createRelease(
            app_name=release['app_name'],
            name=release['name'],
            revision=release['revision'],
            desc=release['desc'])

    yield [_create(RELEASES[0]), _create(RELEASES[1])]


@pytest.mark.vcr()
@pytest.fixture
def packages(server, spc, releases, files):

    def _upload(package, revision):
        pkg = spc.uploadApplicationPackage(
            filepath=package['filepath'],
            app_name=package['app_name'],
            pkg_os=package['os'],
            arch=package['arch'],
            name=package['baseName'],
            repo_type=package['repo_type'],
            repo_url=package['repo_url'],
            revision=revision)
        time.sleep(0.1)
        return pkg

    yield [
        _upload(PACKAGES[0], 'r300'),
        _upload(PACKAGES[1], 'r301'),
        _upload(PACKAGES[2], PACKAGES[2]['revision'])
    ]


@pytest.mark.vcr()
@pytest.fixture
def extensions(server, spc, releases, files):

    def _upload(extension):
        ext = spc.uploadExtension(
            filepath=extension['filepath'],
            app_name=extension['app_name'],
            ext_os=extension['os'],
            arch=extension['arch'],
            name=extension['baseName'],
            repo_type=extension['repo_type'],
            repo_url=extension['repo_url'],
            app_revision=extension['app_revision'],
            revision=extension['revision'])
        time.sleep(0.1)
        return ext

    yield [_upload(EXTENSIONS[0]), _upload(EXTENSIONS[1]), _upload(EXTENSIONS[2])]


@pytest.mark.vcr()
@pytest.fixture(autouse=True)
def TearDown(server, spc, request):
    yield
    for idx in range(len(APPS)):
        try:
            spc.deleteApp(name=APPS[idx])
        except SlicerPackageManagerError:
            pass


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testCreateApp(server, spc):
    app1 = spc.createApp(name=APPS[2])
    assert app1['name'] == APPS[2]

    # Try to create the same application
    with pytest.raises(Exception) as excinfo:
        spc.createApp(name=APPS[2])
    assert 'The Application "%s" already exist.' % APPS[2] in str(excinfo.value)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testListApp(server, spc, apps):
    # List all the app
    app_list = spc.listApp()
    assert len(app_list) == 2
    assert app_list[0]['_id'] == apps[0]['_id']
    assert app_list[1]['_id'] == apps[1]['_id']

    # List one app by name
    app_list = spc.listApp(name=apps[0]['name'])
    assert len(app_list) == 1
    assert app_list[0]['_id'] == apps[0]['_id']


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDeleteApp(server, spc, apps):
    app_list = spc.listApp()
    assert len(app_list) == 2

    deleted_app = spc.deleteApp(name=apps[1]['name'])
    app_list = spc.listApp()
    assert len(app_list) == 1
    assert app_list[0]['_id'] == apps[0]['_id']
    assert deleted_app['_id'] == apps[1]['_id']

    # Try to delete the same deleted app
    with pytest.raises(Exception) as excinfo:
        spc.deleteApp(name=apps[1]['name'])
    assert 'The Application "%s" doesn\'t exist.' % apps[1]['name'] in str(excinfo.value)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testCreateRelease(server, spc, apps):
    release = spc.createRelease(
        app_name=apps[0]['name'], name=RELEASES[0]['name'], revision=RELEASES[0]['revision'])
    assert release['name'] == RELEASES[0]['name']
    assert release['meta']['revision'] == RELEASES[0]['revision']

    # Try to create the same release
    with pytest.raises(Exception) as excinfo:
        spc.createRelease(
            app_name=apps[0]['name'], name=RELEASES[0]['name'], revision=RELEASES[0]['revision'])
    assert 'The release "%s" already exist.' % RELEASES[0]['name'] in str(excinfo.value)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testListRelease(server, spc, apps, releases):
    # List all the releases
    release_list = spc.listRelease(app_name=apps[0]['name'])
    assert len(release_list) == 2
    assert release_list[0]['_id'] == releases[1]['_id']
    assert release_list[1]['_id'] == releases[0]['_id']

    # List one release by name
    release_list = spc.listRelease(app_name=apps[0]['name'], name=releases[0]['name'])
    assert release_list['_id'] == releases[0]['_id']


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDeleteRelease(server, spc, apps, releases):
    release_list = spc.listRelease(app_name=apps[0]['name'])
    assert len(release_list) == 2

    deleted_release = spc.deleteRelease(app_name=apps[0]['name'], name=releases[1]['name'])
    release_list = spc.listRelease(app_name=apps[0]['name'])
    assert len(release_list) == 1
    assert release_list[0]['_id'] == releases[0]['_id']
    assert deleted_release['_id'] == releases[1]['_id']

    # Try to delete the same deleted release
    with pytest.raises(Exception) as excinfo:
        spc.deleteRelease(app_name=apps[0]['name'], name=releases[1]['name'])
    assert 'The release "%s" doesn\'t exist.' % releases[1]['name'] in str(excinfo.value)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testListDraftRelease(server, spc, apps, packages):
    # List all the draft releases
    draft_list = spc.listDraftRelease(app_name=apps[0]['name'])
    assert len(draft_list) == 2

    # List all the draft releases using an offset
    draft_list = spc.listDraftRelease(app_name=apps[0]['name'], offset=1)
    assert len(draft_list) == 1
    assert draft_list[0]['meta']['revision'] == 'r300'

    # List one draft release by revision
    draft_list = spc.listDraftRelease(app_name=apps[0]['name'], revision='r300')
    assert len(draft_list) == 1
    assert draft_list[0]['meta']['revision'] == 'r300'


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testListDraftReleaseWithLimit(server, spc, apps, packages):
    draft_list = spc.listDraftRelease(app_name=apps[0]['name'], limit=1)
    assert len(draft_list) == 1


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDeleteDraftRelease(server, spc, apps, packages):
    draft_list = spc.listDraftRelease(app_name=apps[0]['name'])
    assert len(draft_list) == 2

    deleted_draft = spc.deleteDraftRelease(app_name=apps[0]['name'], revision='r301')
    draft_list = spc.listDraftRelease(app_name=apps[0]['name'])
    assert len(draft_list) == 1
    assert deleted_draft['meta']['revision'] == 'r301'

    # Try to delete the same deleted draft release
    with pytest.raises(Exception) as excinfo:
        spc.deleteDraftRelease(app_name=apps[0]['name'], revision='r301')
    assert 'The release with the revision "%s" doesn\'t exist.' % 'r301' in str(excinfo.value)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testUploadAndDownloadApplicationPackage(server, spc, apps, releases, files):
    # Upload
    pkg1 = spc.uploadApplicationPackage(
        filepath=PACKAGES[0]['filepath'],
        app_name=apps[0]['name'],
        pkg_os=PACKAGES[0]['os'],
        arch=PACKAGES[0]['arch'],
        name=PACKAGES[0]['baseName'],
        repo_type=PACKAGES[0]['repo_type'],
        repo_url=PACKAGES[0]['repo_url'],
        revision=releases[0]['meta']['revision'])
    assert pkg1['meta']['baseName'] == PACKAGES[0]['baseName']

    # Upload
    pkg2 = spc.uploadApplicationPackage(
        filepath=PACKAGES[1]['filepath'],
        app_name=apps[0]['name'],
        pkg_os=PACKAGES[1]['os'],
        arch=PACKAGES[1]['arch'],
        name=PACKAGES[1]['baseName'],
        repo_type=PACKAGES[1]['repo_type'],
        repo_url=PACKAGES[1]['repo_url'],
        revision=releases[0]['meta']['revision'])
    assert pkg2['meta']['baseName'] == PACKAGES[1]['baseName']

    # Download
    downloaded_pkg1 = spc.downloadApplicationPackage(
        app_name=apps[0]['name'], id_or_name=pkg1['_id'])
    downloaded_pkg1_name = apps[0]['meta']['applicationPackageNameTemplate'].format(
        **PACKAGES[0])
    assert downloaded_pkg1['name'] == downloaded_pkg1_name

    # Download
    downloaded_pkg2 = spc.downloadApplicationPackage(
        app_name=apps[0]['name'], id_or_name=pkg2['_id'])
    downloaded_pkg2_name = apps[0]['meta']['applicationPackageNameTemplate'].format(
        **PACKAGES[1])
    assert downloaded_pkg2['name'] == downloaded_pkg2_name

    # Compare downloaded files
    with open('%s.txt' % downloaded_pkg1_name, 'r') as downloaded_f1, open(files[0], 'r') as f1:
        assert downloaded_f1.read() == f1.read()

    with open('%s.txt' % downloaded_pkg2_name, 'r') as downloaded_f2, open(files[1], 'r') as f2:
        assert downloaded_f2.read() == f2.read()

    os.remove('%s.txt' % downloaded_pkg1_name)
    os.remove('%s.txt' % downloaded_pkg2_name)

    # Update an existing package
    spc.uploadApplicationPackage(
        filepath=PACKAGES[1]['filepath'],
        app_name=apps[0]['name'],
        pkg_os=PACKAGES[0]['os'],
        arch=PACKAGES[0]['arch'],
        name=PACKAGES[0]['baseName'],
        repo_type=PACKAGES[0]['repo_type'],
        repo_url=PACKAGES[0]['repo_url'],
        revision=releases[0]['meta']['revision'])
    assert pkg1['meta']['baseName'] == PACKAGES[0]['baseName']

    # Download
    downloaded_pkg1_bis = spc.downloadApplicationPackage(
        app_name=apps[0]['name'], id_or_name=pkg1['_id'])
    downloaded_pkg1_name_bis = apps[0]['meta']['applicationPackageNameTemplate'].format(
        **PACKAGES[0])
    assert downloaded_pkg1_bis['name'] == downloaded_pkg1_name_bis

    # Compare, the file should have changed
    with open('%s.txt' % downloaded_pkg1_name_bis, 'r') as downloaded_f1_bis, open(files[1], 'r') as f1_bis:
        assert downloaded_f1_bis.read() == f1_bis.read()

    os.remove('%s.txt' % downloaded_pkg1_name_bis)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testListApplicationPackage(server, spc, apps, releases, packages):
    # List all the application packages
    pkg_list = spc.listApplicationPackage(app_name=apps[0]['name'])
    assert len(pkg_list) == 3

    # List all the application packages from a release
    pkg_list = spc.listApplicationPackage(app_name=apps[0]['name'], release=RELEASES[0]['name'])
    assert len(pkg_list) == 1
    assert pkg_list[0]['_id'] == packages[2]['_id']

    # List with special filters
    pkg_list = spc.listApplicationPackage(app_name=apps[0]['name'], pkg_os='macosx')
    assert len(pkg_list) == 1
    assert pkg_list[0]['meta']['os'] == 'macosx'


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDeletePackage(server, spc, apps, packages):
    pkg_list = spc.listApplicationPackage(app_name=apps[0]['name'])
    assert len(pkg_list) == 3

    deleted_pkg = spc.deleteApplicationPackage(
        app_name=apps[0]['name'], id_or_name=packages[0]['name'])
    pkg_list = spc.listApplicationPackage(app_name=apps[0]['name'])
    assert len(pkg_list) == 2
    assert deleted_pkg['_id'] == packages[0]['_id']

    # Try to delete the same deleted package
    with pytest.raises(Exception) as excinfo:
        spc.deleteApplicationPackage(
            app_name=apps[0]['name'], id_or_name=packages[0]['name'])
    assert 'The package "%s" doesn\'t exist.' % packages[0]['name'] in str(excinfo.value)


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testUploadAndDownloadExtension(server, spc, apps, releases, files):
    # Upload
    ext1 = spc.uploadExtension(
        filepath=EXTENSIONS[0]['filepath'],
        app_name=apps[0]['name'],
        ext_os=EXTENSIONS[0]['os'],
        arch=EXTENSIONS[0]['arch'],
        name=EXTENSIONS[0]['baseName'],
        repo_type=EXTENSIONS[0]['repo_type'],
        repo_url=EXTENSIONS[0]['repo_url'],
        app_revision=releases[0]['meta']['revision'],
        revision=EXTENSIONS[0]['revision'])
    assert ext1['meta']['baseName'] == EXTENSIONS[0]['baseName']

    # Upload
    ext2 = spc.uploadExtension(
        filepath=EXTENSIONS[1]['filepath'],
        app_name=apps[0]['name'],
        ext_os=EXTENSIONS[1]['os'],
        arch=EXTENSIONS[1]['arch'],
        name=EXTENSIONS[1]['baseName'],
        repo_type=EXTENSIONS[1]['repo_type'],
        repo_url=EXTENSIONS[1]['repo_url'],
        app_revision=releases[0]['meta']['revision'],
        revision=EXTENSIONS[1]['revision'])
    assert ext2['meta']['baseName'] == EXTENSIONS[1]['baseName']

    # Download
    downloaded_ext1 = spc.downloadExtension(
        app_name=apps[0]['name'], id_or_name=ext1['_id'])
    params1 = EXTENSIONS[0].copy()
    params1['app_revision'] = releases[0]['meta']['revision']
    downloaded_ext1_name = apps[0]['meta']['extensionPackageNameTemplate'].format(
        **params1)
    assert downloaded_ext1['name'] == downloaded_ext1_name

    # Download using a different dir_path
    downloaded_ext2 = spc.downloadExtension(
        app_name=apps[0]['name'], id_or_name=ext2['_id'], dir_path='../')
    params2 = EXTENSIONS[1].copy()
    params2['app_revision'] = releases[0]['meta']['revision']
    downloaded_ext2_name = apps[0]['meta']['extensionPackageNameTemplate'].format(
        **params2)
    assert downloaded_ext2['name'] == downloaded_ext2_name

    # Compare downloaded files
    with open('%s.txt' % downloaded_ext1_name, 'r') as downloaded_f1, open(files[0], 'r') as f1:
        assert downloaded_f1.read() == f1.read()

    with open('../%s.txt' % downloaded_ext2_name, 'r') as downloaded_f2, open(files[1], 'r') as f2:
        assert downloaded_f2.read() == f2.read()

    os.remove('%s.txt' % downloaded_ext1_name)
    os.remove('../%s.txt' % downloaded_ext2_name)

    # Update an existing package by changing the revision
    spc.uploadExtension(
        filepath=EXTENSIONS[1]['filepath'],
        app_name=apps[0]['name'],
        ext_os=EXTENSIONS[0]['os'],
        arch=EXTENSIONS[0]['arch'],
        name=EXTENSIONS[0]['baseName'],
        repo_type=EXTENSIONS[0]['repo_type'],
        repo_url=EXTENSIONS[0]['repo_url'],
        app_revision=releases[0]['meta']['revision'],
        revision='newRev')
    assert ext1['meta']['baseName'] == EXTENSIONS[0]['baseName']

    # Download
    downloaded_ext1_bis = spc.downloadExtension(
        app_name=apps[0]['name'], id_or_name=ext1['_id'])
    downloaded_ext1_name_bis = apps[0]['meta']['extensionPackageNameTemplate'].format(
        **params1)
    assert downloaded_ext1_bis['name'] != downloaded_ext1_name_bis

    # Compare, the file should have changed
    with open('%s.txt' % downloaded_ext1_bis['name'], 'r') as downloaded_f1_bis, open(files[1], 'r') as f1_bis:
        assert downloaded_f1_bis.read() == f1_bis.read()

    os.remove('%s.txt' % downloaded_ext1_bis['name'])


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testListExtension(server, spc, apps, extensions):
    # List all the extension packages from draft release
    ext_list = spc.listExtension(app_name=apps[0]['name'])
    assert len(ext_list) == 2

    # List all the extension packages
    ext_list = spc.listExtension(app_name=apps[0]['name'], all=True)
    assert len(ext_list) == 3

    # List all the extension packages from a release
    ext_list = spc.listExtension(app_name=apps[0]['name'], release=RELEASES[0]['name'])
    assert len(ext_list) == 1
    assert ext_list[0]['_id'] == extensions[2]['_id']

    # List with special filters
    ext_list = spc.listExtension(app_name=apps[0]['name'], ext_os='macosx')
    assert len(ext_list) == 1
    assert ext_list[0]['meta']['os'] == 'macosx'


@pytest.mark.vcr()
@pytest.mark.plugin('slicer_package_manager')
def testDeleteExtension(server, spc, apps, extensions):
    ext_list = spc.listExtension(app_name=apps[0]['name'])
    assert len(ext_list) == 2

    deleted_ext = spc.deleteExtension(
        app_name=apps[0]['name'], id_or_name=extensions[0]['name'])
    ext_list = spc.listExtension(app_name=apps[0]['name'])
    assert len(ext_list) == 1
    assert deleted_ext['_id'] == extensions[0]['_id']

    # Try to delete the same deleted extension
    with pytest.raises(Exception) as excinfo:
        spc.deleteExtension(
            app_name=apps[0]['name'], id_or_name=extensions[0]['name'])
    assert 'The extension "%s" doesn\'t exist.' % extensions[0]['name'] in str(excinfo.value)



import os
import pytest


@pytest.fixture
def files():

    filenames = ['file1.txt', 'file2.txt', 'file3.txt', 'file4.txt']

    for idx, filename in enumerate(filenames, start=1):
        with open(filename, 'w+') as file:
            file.write('Content of the file number %s' % idx)

    yield filenames

    for filename in filenames:
        os.remove(filename)



# -*- coding: utf-8 -*-

from girder.models.folder import Folder
from girder.utility.progress import ProgressContext

from . import constants


def isSlicerPackages(item):
    if 'meta' in item and all(k in item['meta'] for k in ('os', 'arch', 'baseName', 'revision')):
        return True
    return False


def getOrCreateReleaseFolder(application, user, app_revision):
    """
    Get or create the release folder associated with the application revision.

    :param application: The parent folder containing the release.
    :param user: The user to check access against or to create the new folder
    :param app_revision: The revision of the application.
    :return: The created/existing release folder.
    """
    release_folder = None
    # Find the release by metadata revision
    releases = Folder().childFolders(application, 'Folder', user=user)
    for folder in releases:
        if 'meta' in folder:
            if 'revision' in folder['meta']:
                if folder['meta']['revision'] == app_revision:
                    release_folder = folder
                    break
    if not release_folder:
        # Only the draft release in the list
        release_folder = list(Folder().childFolders(
            application,
            'Folder',
            user=user,
            filters={'name': constants.DRAFT_RELEASE_NAME}))
        if not release_folder:
            raise Exception('The %s folder not found.' % constants.DRAFT_RELEASE_NAME)
        release_folder = release_folder[0]

        revision_folder = list(Folder().childFolders(
            release_folder,
            'Folder',
            user=user,
            filters={'meta.revision': app_revision}))
        if revision_folder:
            revision_folder = revision_folder[0]
        else:
            revision_folder = Folder().createFolder(
                parent=release_folder,
                name=app_revision,
                parentType='Folder',
                public=release_folder['public'],
                creator=user)
            revision_folder = Folder().setMetadata(
                revision_folder,
                {'revision': app_revision})
        release_folder = revision_folder

    return release_folder


def deleteFolder(folder, progress, user):
    with ProgressContext(progress, user=user,
                         title='Deleting folder %s' % folder['name'],
                         message='Calculating folder size...') as ctx:
        # Don't do the subtree count if we weren't asked for progress
        if progress:
            ctx.update(total=Folder().subtreeCount(folder))
        Folder().remove(folder, progress=ctx)
    return folder
